/*
 * curses.c
 *
 * Curses based IP-Filter logs viewer
 *
 * Copyright (C) 2001  Pierre Berthomier
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2
 * as published by the Free Software Foundation, provided that this notice
 * is preserved in its entirety and due credit is given to the original
 * author and the contributors.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#include "vlog.h"

void init_curses()
{
  mode_curses = 1;
  initscr();
  // start_color();
  raw();
  cbreak();
  noecho();
  nonl();
  intrflush(stdscr, FALSE);
  keypad(stdscr, TRUE);		/* cook escape sequences keys into one char */
  nodelay(stdscr, TRUE);	/* make getch() non-blocking) */

  getmaxyx(stdscr, screen_maxy, screen_maxx);
  if (screen_maxy < 5) {
    err("Terminal is not tall enough.");
  }
  logspad_nlines = screen_maxy - 2;

  create_windows();
}

void create_windows()
{
  if (headpad) {
    delwin(headpad);	// in case of window resize, these windows already exist
    delwin(logspad);	// => first destroy them
    delwin(statswin);
  }

  headpad  = newpad(1, PADWIDTH);			// ipf fields headers
  logspad  = newpad(logspad_nlines, PADWIDTH);		// middle window for log lines
  statswin = newwin(1, screen_maxx, screen_maxy - 1, 0);// bottom status window

  leaveok(headpad,  TRUE);
  leaveok(logspad,  TRUE);
  leaveok(statswin, TRUE);

  clearok(stdscr, TRUE);
  refresh();
  clearok(stdscr, FALSE);
}

void sigwinch_handler(int signal)
{
  screen_was_resized = 1;
}

void init_signal_handlers()
{
  (void) signal(SIGWINCH, sigwinch_handler);
}

/* Return the length of the chtype * string passed */

int chstrlen(chtype * chstr)
{
  int i = 0;

  if (chstr)
    while (*chstr++)
      i++;
  return i;
}


/* Given a string (char * format and optional arguments), convert it
 * to a chstr * (chars + attributes).
 * - if chstr is not null, it will receive the converted str
 * - if chstr is null, this routine allocates it.
 * All characters will have the attribute 'attr' (A_NORMAL for none).
 */

chtype * str_to_chstr(chtype * chstr, int attr, char * format, ...)
{
  va_list  arglist;
  char     str[BUFSIZ];
  char   * s;
  chtype * cs;

  va_start(arglist, format);		// prepare str with the format strings and args
  vsprintf(str, format, arglist);
  va_end(arglist);

  if (! chstr)		
    chstr = (chtype *) tmalloc(sizeof(chtype) * (1 + strlen(str)));

  for (s = str, cs = chstr; *s; )	// copy str to chstr with attribute attr
    *cs++ = (int) *s++ | attr;
  *cs = 0;

  return chstr;
}

/* Append a normal string (format + optional args) to an existing (chtype *) string,
 * with video attribute attr. chstr is supposed to be long enough
 * to receive str at its end.
 */

chtype * strcat_to_chstr(chtype * chstr, int attr, char * format, ...)
{
  va_list  arglist;
  char     str[BUFSIZ];
  char   * s;
  chtype * cs;

  va_start(arglist, format);		// prepare str with the format strings and args
  vsprintf(str, format, arglist);
  va_end(arglist);

  for (cs = chstr; *cs; cs++);		// advance to end of chstr

  for (s = str; *s; )			// copy str to chstr with attribute attr
    *cs++ = (int) *s++ | attr;
  *cs = (int) 0;

  return chstr;
}

/* Allocate a 'length' chars string and fill it with 'myattr' */

char * create_const_attrline(int length, char myattr)
{
  char * s;
  int    i;
  char * str = (char *) tmalloc(length + 1);

  for (i = 0, s = str; i < length; i++)
    *s++ = myattr;
  *s = '\0';

  return str;
}


void transient_message(char * format, ...)
{
  va_list  arglist;
  char     str[PADWIDTH + 1];
  chtype   chstr[PADWIDTH + 1];

  va_start(arglist, format);
  vsprintf(str, format, arglist);
  va_end(arglist);

  if (mode_curses) {
    str_to_chstr(chstr, A_NORMAL, "%s", str);

    clear_status_window();
    mvwaddchstr(statswin, 0, 0, chstr);
    wnoutrefresh(statswin);
    doupdate();
    sleep(2);
    draw_status_window();
    refresh_screen();

  } else
    fprintf(stderr, "%s: %s\n", MYNAME, str);
}

/* Given a logline, fill in the passed buffer with ASCII chars and
 * eventually ANSI codes to render bold, underlined and reverse chars.
 */

char * fmt_and_attr_lines_to_ANSI_string(
  logline * ll, char * buf)
{
  char * s;
  char * t;
  char * b = buf;
  chtype last_attr = 0;

  s = ll->fmtline;
  t = ll->attrline;
  while (*s) {
    if (*t != last_attr) {					// attribute change:
      strcat_and_advance_pointer(b, ANSI_NORMAL);		// insert ANSI codes in buf
      if (*t & MYA_BOLD)	strcat_and_advance_pointer(b, ANSI_BOLD);
      if (*t & MYA_REVERSE)	strcat_and_advance_pointer(b, ANSI_REVERSE);
      if (*t & MYA_UNDERLINE)	strcat_and_advance_pointer(b, ANSI_UNDERLINE);
    }

    *b++ = *s++;		// add this char
    last_attr = *t++;		// remember new 'last-attr'
  }

  strcat_and_advance_pointer(b, ANSI_NORMAL);
  *b = '\0';
  return buf;
}


/* -------------------------------------------------------------------------- */

void help_screen ()
{
  char   helpbuf[65536];
  chtype chstr[PADWIDTH];
  char * s;
  char * t;
  int    row;

  sprintf(helpbuf, "
    %s-%s (%s)
    Logfile viewer for IP-Filter       ----- JUMP KEYS ---------------------    
    %-32s     g   go to a given line
                                         G   goto end or return to prev. pos
--------- MOVEMENT KEYS ------------     T   goto top
 arrows   move the window one line       0   scroll window to leftmost
 h,j,k,l  in either direction.         
 H,J,K,L  Caps move faster.             /,?  search a regexp forward/backward
 b,space  back, forward 1 page          n,N  goto next/previous occurence
PgUp/Down these work on XFree only      >,<  goto next/previous day boundary
                                        ],[  goto next/previous marked line
--------- MISC KEYS ----------------     
    a     toggle AutoScroll mode        s,S  switch to another logfile
    r     toggle raw/formatted  mode     &   display hostname/logfile
    i     ip address display style       !   spawn a subshell
    m     mark (reverse) bottom line   
    M     clear all marks                
   ^L     redraw screen
%s
    Q     quit
",
MYNAME, VERSION, DATE, CREDIT,
  (mode_ipmon
  ? "    o     toggle display options  (d=date, t=time, u=microsec, h=hostname,
          f=intf, r=ipf rule numbers, i=isba rule numbers, l=ip header length)"
  : (mode_httpd_access
     ? "    o     toggle display options (d=date, t=time, c=client IP, u=user)"
     : (mode_httpd_errors
        ? "    o     toggle display options (d=date, t=time)"
	: "    o     toggle display options (d=date, t=time, h=host, p=proc. name, i=pid)"
))));

  row = 0;
  s = helpbuf;

  while (row < logspad_nlines) {
    for (t = s; *t && (*t != '\n'); )		// t -> next newline
      t++;

    if (*t)
      *t = '\0';				// newline found, replace with EOL
    else
      t = NULL;					// newline not found. end of text.

    mvwaddchstr(logspad, row, 0,		// draw one line
      str_to_chstr(chstr, A_NORMAL, "%-*s", PADWIDTH - 1, (t ? s : "")));

    if (t)
      s = ++t;
    row++;
  }

  clearok(stdscr, 1);				// update screen
  pnoutrefresh(logspad, 0, 0,  1, 0, 1 + (logspad_nlines - 1), screen_maxx);
  doupdate();
  clearok(stdscr, 0);

  nodelay(stdscr, FALSE);			// make getch() blocking
  getch();  					// and wait for a key
  nodelay(stdscr, TRUE);
  hard_redraw_screen();
}


void draw_head_pad()
{
  chtype chstr[PADWIDTH + 1];

  printd1("draw_head_pad(): mode_ipmon=%d.\n", mode_ipmon);

  if (mode_raw_cooked) {
    if (mode_ipmon)
      ipf_headers_line(chstr);
    else if (mode_httpd_access || mode_httpd_errors)
      httpd_headers_line(chstr);
    else
      syslog_headers_line(chstr);

  } else {
    str_to_chstr(chstr, A_REVERSE, "%s ", "Raw lines ");
    headers_append_host_logfilename(chstr);
    strcat_to_chstr(chstr, A_REVERSE, "%*s", PADWIDTH - chstrlen(chstr) - 1, "");
  }

  mvwaddchstr(headpad, 0, 0, chstr);
}


chtype * headers_append_host_logfilename(
  chtype * chstr)
{
  int    i;
  char * s;

  i = screen_maxx - chstrlen(chstr) - strlen(logfile) - strlen(local_host_name) - 2;
  if (i > 2) {								// try to display hostname:/path/to/logfile
    strcat_to_chstr(chstr, A_REVERSE, "%*s%s:%s",			// in the headers line
		    i, "", local_host_name, logfile);
  } else {
    i = screen_maxx - chstrlen(chstr) - strlen(logfile) - 1;		// if there's not enough room, try to display
    if (i > 2) {							// only /path/to/logfile
      strcat_to_chstr(chstr, A_REVERSE, "%*s%s", i, "", logfile);
    
    } else {								// otherwise try to display only the basename
      s = basename(logfile);
      i = screen_maxx - chstrlen(chstr) - strlen(s) - 1;
      if (i > 2) {
        strcat_to_chstr(chstr, A_REVERSE, "%*s%s", i, "", s);
      }
    }
  }

  return chstr;
}


/* log_line_to_fmt_and_attr_lines(): compute ll->fmtline and ll->attrline
 * Formatted line chars are stored in ll->fmtline,
 * attributes (bold, inverse video, underline) are stored in ll->attrline.
 * Use the fields computed (or compute and store them in ll->fields)
 */

void log_line_to_fmt_and_attr_lines(
  logline * ll)
{
  int make_it_raw = 1;
  int make_it_raw_attr;

  make_it_raw_attr = (ll->is_marked ? MYA_REVERSE : MYA_NORMAL) | MY_BIT7;

  if (ll->is_a_separator)
    make_it_raw_attr = ((ll->is_marked ? MYA_REVERSE : MYA_NORMAL) ^ MYA_REVERSE) | MY_BIT7;

  else if (mode_raw_cooked) {				// COOKED MODE ? go compute what to display
    if (mode_ipmon) {					// if ipmon mode:
      if (ll->not_an_ipmon_line)			// if it's not an ipmon line, display as raw
        make_it_raw_attr = (ll->is_marked ? MYA_REVERSE : MYA_NORMAL) | MY_BIT7;
      else {
        if (! ll->fields)				// if it's an ipmon line, cut the log lines
          cut_logline_to_ipfilter_fields(ll);		// into fields if not already done

	if (ll->fields->unknown_format)			// if format not recognised: raw, video reverse
          make_it_raw_attr = ((ll->is_marked ? MYA_REVERSE : MYA_NORMAL) ^ MYA_REVERSE) | MY_BIT7;

	else {						// if ipmon format was recognised,
          ipmon_fields_to_fmt_and_attr_lines(ll);	// prepare the formatted line from computed ipmon fields
          make_it_raw = 0;
        }
      }
    } else if (mode_httpd_access || mode_httpd_errors) {
      httpd_fields_to_fmt_and_attr_lines(ll);		// prepare the formatted line from computed httpd fields
      make_it_raw = 0;

    } else {
      syslog_fields_to_fmt_and_attr_lines(ll);		// prepare the formatted line from computed syslog fields
      make_it_raw = 0;
    }
  }

  if (make_it_raw) {					// RAW MODE
    if (ll->fmtline && ll->fmtline != ll->line)
      tfree(ll->fmtline);
    if (ll->attrline)
      tfree(ll->attrline);

    ll->fmtline = ll->line;
    ll->attrline = create_const_attrline(strlen(ll->line), make_it_raw_attr);
  }

  ll->serial = cur_fmt_serial;				// remember this logline preformatting is up-to-date
}


void draw_loglines_pad()
{
  logline * ll;
  int       row = logspad_nlines - 1;
  chtype    chstr[PADWIDTH + 1];
  chtype *  cs;
  char *    s;
  u_char *  t;
  chtype    ch, attr;
  int       i;
  char      cur_month[SMALLSTR];
  int	    cur_day = 0;
  int       match_mask;

  if ((ll = cur_bottom_ll)) {

    while (ll && (row >= 0)) {

      if ((ll->serial != cur_fmt_serial) || (! ll->fmtline))
        log_line_to_fmt_and_attr_lines(ll);	// go compute what to display if not done

      attr = A_NORMAL;
      if (ll->day && (ll->day != cur_day || strncmp(ll->month, cur_month, 3))) {
        if (cur_day)				// if day changed since last logline,
	  attr = A_UNDERLINE;			// underline this one
        cur_day = ll->day;
        strncpy(cur_month, ll->month, 3);
      }

      // if this line is out of date with respect to current search regexp, ignore the SEARCH_MATCH bit
      match_mask = (ll->srch_serial == cur_srch_serial ? 0xff : ~MY_SEARCH_MATCH);

      s  = ll->fmtline;
      t  = ll->attrline;
      cs = chstr;
      i  = 0;
      while (*s) {				// build chtype * string
        *cs++ = (chtype) *s++ | attr | my_attr_table[(int) *t++ & match_mask];
	i++;
      }

      ch = (chtype) ' ' | attr | ((ll->is_marked ^ ll->is_a_separator) ? A_REVERSE : A_NORMAL);

      while (i++ < PADWIDTH)			// pad chstring -> PADWIDTH
	*cs++ = ch;
      *cs = '\0';

      mvwaddchstr(logspad, row, 0, chstr);	// send it to screen buffer
      ll = ll->prev;
      row--;
    }

    while (row-- >= 0)				// if beginning of log: blank upper lines
      mvwaddchstr(logspad, row, 0, str_to_chstr(chstr, A_NORMAL, "%*s", PADWIDTH, ""));
  }
}


/*
 * Draw status line
 */

void draw_status_window()
{
  chtype status[PADWIDTH];
  char   buf[SMALLSTR];
  char   position[SMALLSTR];

  if ((search_status & SRCH_IN_PROGRESS) && search_nbr_matches)
    sprintf(buf, "%d matches", search_nbr_matches);
  
  *position = '\0';
  if (i_am_at_top)			// compute position (top, bottom, or xx%)
    strcpy(position, "top of");

  else if (cur_bottom_ll == tail_line)
    strcpy(position, "bottom of");

  else if (cur_bottom_ll && head_line) {
    sprintf(position, "%3d%% of", (int)
     ((float) 99.5 * (cur_bottom_ll->linenum - head_line->linenum) / total_nbr_loglines_read));
  }

  str_to_chstr(status, A_REVERSE, "%s-%s ", MYNAME, VERSION);
  strcat_to_chstr(status, A_REVERSE, "%3s %10s  %9s %d lines [%s]  %s ",
    (mode_raw_cooked ? ""            : "Raw"),
    (mode_autoscroll ? "AutoScroll" : ""),
    position,
    total_nbr_loglines_read, 
    ((search_status & SRCH_IN_PROGRESS)
     ? (search_nbr_matches ? buf : "Searching")
     : reverse_thread_status),
     search_status & SRCH_IN_PROGRESS ? "ESC = cancel search" : "* = help");

  localtime_to_buf(buf, "%H:%M:%S");
  strcat_to_chstr(status, A_REVERSE, " %s ", buf);

  strcat_to_chstr(status, A_REVERSE, "%*s", PADWIDTH - chstrlen(status) - 1, "");
  mvwaddchstr(statswin, 0, 0, status);
}


void clear_status_window()
{
  chtype chstr[PADWIDTH];
  
  mvwaddchstr(statswin, 0, 0, str_to_chstr(chstr, A_NORMAL, "%-*s", PADWIDTH - 1, ""));
  wnoutrefresh(statswin);
  doupdate();
}


void refresh_screen()
{
  clearok(stdscr, 1);
  move(0, screen_maxx - 1);
  pnoutrefresh(headpad, 0, horiz_scroll,   0, 0, 0, screen_maxx);
  pnoutrefresh(logspad, 0, horiz_scroll,   1, 0, 1 + (logspad_nlines - 1), screen_maxx);
  wnoutrefresh(statswin);

  doupdate();
  clearok(stdscr, 0);
}

void redraw_screen()
{
/*
  redrawwin(headpad);	//// dumps core in OpenBSD
  redrawwin(logspad);
  redrawwin(statswin);
*/
  draw_head_pad();
  draw_loglines_pad();
  draw_status_window();

  touchwin(headpad);
  touchwin(logspad);
  touchwin(statswin);

  refresh_screen();
}

void redraw_status_window()
{
  draw_status_window();
  clearok(stdscr, 1);
  wnoutrefresh(statswin);
  doupdate();
  move(0, screen_maxx - 1);
  clearok(stdscr, 0);
}

void hard_redraw_screen()			// called after a window resize
{
  endwin();					// one cycle endwin/init_curses is enough on Solaris,
  init_curses();				// but OpenBSD requires several
  endwin();
  init_curses();
  redraw_screen();
  init_signal_handlers();
  screen_was_resized = 0;
}

void clean_curses_and_exit(int return_code)
{
  clear_status_window();
  nocbreak();
  flushinp();		// clear keyboard buffer
  endwin();
  close_debugfile();
  exit(return_code);
}


// --------------------------------------------------------------------------------------

int search_logline(
  logline * ll)
{
  char * s;
  char * t;
  int    found;
  int    len;
  int    start;
  regmatch_t pmatch[32];

  if ((ll->serial != cur_fmt_serial) || (! ll->fmtline)) // update the formatted line if necessary
    log_line_to_fmt_and_attr_lines(ll);			 // (if user changed format for example)

  else if (ll->srch_serial == cur_srch_serial)		 // if this line has already been searched
    return ll->matched;					 // with current regexp, return result

  else {
    for (t = ll->attrline; *t; t++)			// cleanup any previous matched substring highlight
      *t &= ~MY_SEARCH_MATCH;
    ll->matched = 0;					// clear this logline match flag
  }

  found = 0;
  start = 0;
  s = ll->fmtline;
  while (search_regexp					// (in case this is set to NULL)
      && ! regexec(search_regexp, s, 31, pmatch, 0)) {	// using a while loop allows
    len = pmatch[0].rm_eo - pmatch[0].rm_so;		// multiple occurences in the same line
    if (len > 0) {
      for (t = (char *) (ll->attrline + start + pmatch[0].rm_so) ; len > 0; len--)
        *t++ |= MY_SEARCH_MATCH;			// highlight matched substrings
    }
    ll->matched = 1;					// flag this line as `matched'
    found = 1;

    s = (char *) (s + pmatch[0].rm_so + 1);		// now examine next part of string
    start += pmatch[0].rm_so + 1;
  }

  ll->srch_serial = cur_srch_serial;			// remember this line is up-to-date
							// (any matches will be highlighted)
  return found;
}

void search_thread(
  int c)
{
  logline *  start_ll;
  logline *  ll;
  int        backwards;
  int        i;

  backwards		= (c == '?' ? 1 : 0);
  search_result_ll	= (logline *) NULL;
  search_nbr_matches	= 0;

  ll = cur_bottom_ll;
  i  = 0;
  while (ll->prev && i++ < logspad_nlines) {	// if there are any matches in displayed
    search_logline(ll);				// lines, highlight them immediately.
    ll = ll->prev;
  }
  draw_loglines_pad();
  refresh_screen();

  ll = cur_bottom_ll;				// then start search from middle of screen
  ll_walk_back(ll, logspad_nlines / 2 + backwards);
  start_ll = ll;

  debug_int = 0;
  while (! search_thread_stop) {		 // is like a while (true) except that
    if (backwards)				 // main thread may set search_thread_stop to 1
         ll = (ll->prev ? ll->prev : tail_line); // logline queue wrap around
    else ll = (ll->next ? ll->next : head_line);

    debug_int++;

    if (search_logline(ll)) {			// search regexp in logline

      search_nbr_matches++;			// count matching lines
      if (! search_result_ll) {
	search_result_ll = ll;			// notify main thread that I've found 1 match
        search_status |= SRCH_FOUND_ONE;
      }
    }
    if (ll == start_ll)				// cycled through all queue ?
      break;
  }

  search_status &= ~SRCH_IN_PROGRESS;		// search is not running any more
  search_status |= SRCH_ENDED;			// notify that search thread ended
  search_thread_stop = 0;
}


void kill_search_thread()
{
  search_thread_stop = 1;
  napms(252);
  search_status = SRCH_NOP;
  search_regexp = NULL;
  cur_srch_serial++;
  redraw_status_window();
}


void get_search_regexp(int forw_backw)
{
  chtype    chstr[PADWIDTH + 1];
  char      regexp_str[BUFSIZ];
  regex_t * re;
  int       ret;

  clear_status_window();
  str_to_chstr(chstr, A_NORMAL, "Search ");
  strcat_to_chstr(chstr, A_NORMAL, (forw_backw ? "backward: " : "forward: "));
  mvwaddchstr(statswin, 0, 0, chstr);
  wnoutrefresh(statswin);
  doupdate();

  move(screen_maxy - 1, chstrlen(chstr));
  echo();
  nodelay(stdscr, FALSE);	// make getch() blocking
  getstr(regexp_str);

  if (strlen(regexp_str)) {
    re = (regex_t *) tmalloc(sizeof(regex_t));
    ret = regcomp(re, regexp_str, REG_EXTENDED);
    if (ret) {
      search_regexp = (regex_t *) NULL;
      transient_message("regexp is malformed (code=%d)", ret);
    } else {
      search_regexp = re;
      cur_srch_serial++;
    }
  } else {
    search_regexp = (regex_t *) NULL;
    cur_srch_serial++;
    draw_loglines_pad();
  }

  nodelay(stdscr, TRUE);	// make getch() non-blocking
  noecho();
  draw_status_window();
  refresh_screen();
}


// --------------------------------------------------------------------------------------


void goto_line_percent()
{
  chtype     chstr[PADWIDTH + 1];
  char       str[BUFSIZ];
  int        percent;
  int        linecount;
  logline * ll;

  clear_status_window();
  *chstr = (chtype) 0;
  strcat_to_chstr(chstr, A_NORMAL,  "go to line (percent) (0-100) : ");
  mvwaddchstr(statswin, 0, 0, chstr);
  wnoutrefresh(statswin);
  move(screen_maxy - 1, chstrlen(chstr));
  doupdate();

  move(screen_maxy - 1, chstrlen(chstr));
  echo();
  nodelay(stdscr, FALSE);	// make getch() blocking
  getstr(str);

  if (*str && isdigit((int) *str)) {
    sscanf(str, "%d", &percent);

    if (percent == 0) {
      cur_bottom_ll = head_line;
      ll_walk_forward(cur_bottom_ll, logspad_nlines - 1);	// ensure screen will be full of lines
      i_am_at_top = 1;

    } else if (percent == 100)
      cur_bottom_ll = tail_line;

    else if (percent >= 0 && percent <= 100) {
      linecount = (int) ((float) (99.5 - percent) * total_nbr_loglines_read / 100);
      ll = tail_line;
      while (linecount-- && ll->prev)
        ll = ll->prev;

      ll_walk_back(ll, logspad_nlines);				// ensure screen will be full of lines
      if (ll->prev) {
        ll_walk_forward(ll, logspad_nlines - 1);
      } else {
        ll_walk_forward(ll, logspad_nlines - 1);
	i_am_at_top = 1;
      }
      cur_bottom_ll = ll;
      
    }
  }

  nodelay(stdscr, TRUE);	// make getch() non-blocking
  noecho();
  draw_status_window();
  redraw_screen();
}


void load_other_logfile(
  int key_or_fullname)
{
  chtype chstr[PADWIDTH + 1];
  char   str[BUFSIZ];
  char * logfile;
  int    i;
  char * pstr;
  FILE * fd;

  clear_status_window();

  *chstr = (chtype) 0;
  if (key_or_fullname == 1)
    strcat_to_chstr(chstr, A_NORMAL, "other logfile pathname: ");

  else {
    pstr = str;
    for (i = 0; i < 256; i++)
      if (key_to_logfile[i])
        *pstr++ = (char) i;
    *pstr = '\0';
    strcat_to_chstr(chstr, A_NORMAL, "other logfile [%s]: ", str);
  }

  mvwaddchstr(statswin, 0, 0, chstr);
  wnoutrefresh(statswin);
  move(screen_maxy - 1, chstrlen(chstr));
  doupdate();
  echo();
  nodelay(stdscr, FALSE);	// make getch() blocking

  if (key_or_fullname == 1)
    getstr(str);
  else {
    *str     = getch();
    if (*str == ERR || *str == '\r' || *str == (char) 27 || *str == ' ')   // Return, Space, ESC: cancel
      *str     = '\0';
    else
      *(str+1) = '\0';
  }

  logfile = (char *) NULL;

  if (*str) {
    if (strlen(str) > 1)
      logfile = (char *) strdup(str);

    else if (! (logfile = key_to_logfile[(int) *str]))
      transient_message("that's not a predefined logfile path I know of ...");

    if (logfile) {
      // fprintf(stderr, "MYPATHNAME = %s, logfile = %s \n", mypathname, logfile);

      if ((fd = fopen(logfile, "r"))) {
	nocbreak();
	flushinp();
	endwin();
	printd2("execlp(%s, vlog, -l, %s, NULL)\n", mypathname, logfile);
        execlp(mypathname, "vlog", "-l", logfile, (char *) NULL);
	init_curses();
        transient_message("couldn't relaunch myself: %s (%d)", strerror(errno), errno);
      } else {
        transient_message("can't read %s: %s (%d)", logfile, strerror(errno), errno);
      }
    }
  }

  nodelay(stdscr, TRUE);	// make getch() non-blocking
  noecho();
  draw_status_window();
  refresh_screen();
}


void get_what_to_toggle()
{
  chtype chstr[PADWIDTH + 1];
  int c;

  clear_status_window();

  *chstr = (chtype) 0;
  if (screen_maxx > 98 || ! mode_ipmon)
    strcat_to_chstr(chstr, A_NORMAL, "display ");
  if (screen_maxx > 90 || ! mode_ipmon)
    strcat_to_chstr(chstr, A_NORMAL, "options: ");
  strcat_to_chstr(chstr,   (option_show_date        ? A_BOLD : A_NORMAL), "[d]ate ");
  strcat_to_chstr(chstr,   (option_show_time        ? A_BOLD : A_NORMAL), "[t]ime ");

  if (mode_ipmon)
    strcat_to_chstr(chstr, (option_show_microsec    ? A_BOLD : A_NORMAL), "[u]sec ");
  if (mode_ipmon || (! mode_httpd_access && ! mode_httpd_errors))
    strcat_to_chstr(chstr,   (option_show_hostname    ? A_BOLD : A_NORMAL), "[h]ost ");

  if (mode_ipmon) {
    strcat_to_chstr(chstr, (option_show_intf        ? A_BOLD : A_NORMAL), "int[f] ");
    strcat_to_chstr(chstr, (option_show_rulenum     ? A_BOLD : A_NORMAL), "[r]ule ");
    if (isba_rulenums_available) {
      strcat_to_chstr(chstr, (option_show_isbanum   ? A_BOLD : A_NORMAL), "[i]sba ");
      if (screen_maxx > 90)
        strcat_to_chstr(chstr, (option_show_isbanum ? A_BOLD : A_NORMAL), "rule  ");
    }
    strcat_to_chstr(chstr, (option_show_headerlen   ? A_BOLD : A_NORMAL), "hdr[l]en ");

  } else {
    if (mode_httpd_access) {
      strcat_to_chstr(chstr, (option_show_httpd_client_ip   ? A_BOLD : A_NORMAL), "[s]rc addr ");
      strcat_to_chstr(chstr, (option_show_httpd_client_user ? A_BOLD : A_NORMAL), "[u]ser ");
      strcat_to_chstr(chstr, (option_show_httpd_code        ? A_BOLD : A_NORMAL), "[c]ode ");
      strcat_to_chstr(chstr, (option_show_httpd_length      ? A_BOLD : A_NORMAL), "[l]en ");
    } else if (! mode_httpd_errors) {
      strcat_to_chstr(chstr, (option_show_proc_name   ? A_BOLD : A_NORMAL), "[p]rocess name ");
      strcat_to_chstr(chstr, (option_show_pid         ? A_BOLD : A_NORMAL), "p[i]d ");
    }
  }

  strcat_to_chstr(chstr, A_NORMAL,  "or [a]ll/[n]one :");
  mvwaddchstr(statswin, 0, 0, chstr);
  wnoutrefresh(statswin);
  move(screen_maxy - 1, chstrlen(chstr));
  doupdate();

  nodelay(stdscr, FALSE);	// make getch() blocking

  cur_fmt_serial++;
  if ((c = getch()) != ERR) {
    switch (c) {
      case 'd': option_show_date		= 1 - option_show_date;      break;
      case 't': option_show_time		= 1 - option_show_time;      break;
      case 'u': if (mode_ipmon)
		  option_show_microsec		= 1 - option_show_microsec;
		else
                  option_show_httpd_client_user	= 1 - option_show_httpd_client_user;
		break;
      case 'h': option_show_hostname		= 1 - option_show_hostname;  break;
      case 'f': option_show_intf		= 1 - option_show_intf;      break;
      case 'r': option_show_rulenum		= 1 - option_show_rulenum;   break;
      case 'i': if (mode_ipmon)
		  option_show_isbanum		= 1 - option_show_isbanum;
		else
		  option_show_pid		= 1 - option_show_pid;
		break;
      case 'l': if (mode_ipmon)
		  option_show_headerlen		= 1 - option_show_headerlen;
		else
                  option_show_httpd_length	= 1 - option_show_httpd_length;
		break;
      case 'p': option_show_proc_name		= 1 - option_show_proc_name; break;
      case 's': option_show_httpd_client_ip	= 1 - option_show_httpd_client_ip; break;
      case 'c': option_show_httpd_code		= 1 - option_show_httpd_code; break;
      case 'a': option_show_date        = option_show_time      = option_show_microsec =
      		  option_show_hostname  = option_show_intf      = option_show_rulenum =
		  option_show_isbanum   = option_show_headerlen =
		  option_show_proc_name = option_show_pid       =
		  option_show_httpd_client_ip = option_show_httpd_client_user =
		  option_show_httpd_code = option_show_httpd_length = 1;
      		break;
      case 'n': option_show_date        = option_show_time      = option_show_microsec =
      		  option_show_hostname  = option_show_intf      = option_show_rulenum =
		  option_show_isbanum   = option_show_headerlen =
		  option_show_proc_name = option_show_pid       =
		  option_show_httpd_client_ip = option_show_httpd_client_user =
		  option_show_httpd_code = option_show_httpd_length = 0;
      		break;
      default:  cur_fmt_serial--;
    }
  }

  nodelay(stdscr, TRUE);	// make getch() non-blocking
}


void scroll_screen(int direction, int how_much)
{
  int n;
  logline * ll;

  switch (direction) {
    case SCROLL_DOWN:
      if (cur_bottom_ll) {
        while (how_much-- > 0 && cur_bottom_ll->next)	// go forward 'how_much' lines
          cur_bottom_ll = cur_bottom_ll->next;
        i_am_at_top = 0;
      }
      break;

    case SCROLL_RIGHT:
      horiz_scroll = min(PADWIDTH - screen_maxx - 1, horiz_scroll + how_much);
      break;

    case SCROLL_UP:
      if (cur_bottom_ll) {
        while ((how_much-- > 0) && cur_bottom_ll->prev)	// go back 'how_much' lines
          cur_bottom_ll = cur_bottom_ll->prev;

        n = logspad_nlines;				// go back one screen to see
        ll = cur_bottom_ll;				// if the screen is full
        while (n-- > 0 && ll->prev)
          ll = ll->prev;

        i_am_at_top = (n >= 0);

        while (n-- > 0 && cur_bottom_ll->next)		// adjust cur_bottom_ll to have
          cur_bottom_ll = cur_bottom_ll->next;		// a screen full of lines

      }
      break;

    case SCROLL_LEFT:
      horiz_scroll = max(0, horiz_scroll - how_much);
      break;
  }

  draw_status_window();
  draw_loglines_pad();
  refresh_screen();
}

void go_to_top_of_logs()
{
  if ((cur_bottom_ll = head_line)) {
    i_am_at_top = 1;
    if (mode_curses) {
      ll_walk_forward(cur_bottom_ll, logspad_nlines - 1);
      redraw_screen();
    }
  }
}


logline * center_logline_in_screen(
  logline * ll)
{
  logline * origll = ll;

  ll_walk_back(ll, logspad_nlines / 2);
  if (ll->prev) {				// there are enough lines to fill top of screen
    ll_walk_forward(origll, logspad_nlines / 2);
    return origll;

  } else {					// there are not enough lines to fill top
    ll_walk_forward(ll, logspad_nlines - 1);	// => go forward 1 fuill screen
    i_am_at_top = 1;
    return ll;
  }
}


// --------------------------------------------------------------------------------------

void react_to_typed_keys()
{
  int       c, i;
  logline * ll;
  logline * start_ll;
  char      cur_month[SMALLSTR];
  int	    cur_day;
  int	    ret;
  int	    backwards;

  while ((c = getch()) != ERR) {
    last_typed_key_second = current_second();

    switch (c) {
    case 0x0c:					// '^L' = redraw screen
      hard_redraw_screen();
      break;

    case 'b':					// 'b' = back one page
    case 0x153:					// PgUp (XFree)
    case 0x02:					// Ctrl-B
      scroll_screen(SCROLL_UP, logspad_nlines); break;

    case ' ':					// ' ' = forward one page
    case 0x152:					// PgDown (XFree)
    case 0x06:					// Ctrl-F
      scroll_screen(SCROLL_DOWN, logspad_nlines); break;

    case 'K':					// K, Ctrl-U = back half a page
    case 0x15:
      scroll_screen(SCROLL_UP, logspad_nlines / 2); break;

    case 'J':					// J, Ctrl-D = forward half a page
    case 0x04:
      scroll_screen(SCROLL_DOWN, logspad_nlines / 2); break;

    case KEY_RIGHT:				// ->, l = scroll left 1 char
    case 'l':
      scroll_screen(SCROLL_RIGHT, 1);	break;

    case KEY_LEFT:				// <-, h = scroll right 1 char
    case 'h':
      scroll_screen(SCROLL_LEFT, 1);	break;

    case KEY_UP:				// KEY_UP, k = scroll 1 line down
    case 'k':
      scroll_screen(SCROLL_UP, 1);	break;

    case KEY_DOWN:				// KEY_DOWN, j, RETURN = scroll 1 line up
    case 'j':
    case 0x0d:
      scroll_screen(SCROLL_DOWN, 1);	break;

    case 'L':					// L = scroll left 30 chars
      scroll_screen(SCROLL_RIGHT, 30);	break;

    case 'H':					// H = scroll right 30 chars
      scroll_screen(SCROLL_LEFT, 30);	break;

    case '^':					// ^ or 0 = scroll back to leftmost
    case '0':
      horiz_scroll = 0;
      redraw_screen();
      break;

    case 'g':					// g = type line to go to (in percent: 0-100)
      goto_line_percent();
      break;

    case 'G':					// G = go to bottom *or* go back to where we were
    case KEY_END:				// or End (XFree)
      if (cur_bottom_ll != tail_line) {
        last_ll = cur_bottom_ll;
        cur_bottom_ll = tail_line;
      } else
        if (last_ll)
	  cur_bottom_ll = last_ll;

      i_am_at_top = 0;
      redraw_screen();
      break;

    case 'T':					// T = go to top (1st logline)
    case KEY_HOME:				// of Home (XFree)
      go_to_top_of_logs();
      break;

    case 'r':					// r = toggle display mode: cooked/raw loglines
   // if (option_z_compress)
   //   transient_message("raw mode not available whith option -z");
   // else {
        mode_raw_cooked = 1 - mode_raw_cooked;
        cur_fmt_serial++;
        redraw_screen();
   // }
      break;

    case 'a':					// a = toggle auto-goto-bottom mode
      mode_autoscroll = 1 - mode_autoscroll;
      redraw_status_window();
      break;

    case 'i':					// i = toggle sparse/compact IP addresses
      option_sparse_ip_addr = 1 - option_sparse_ip_addr;
      cur_fmt_serial++;
      redraw_screen();
      break;

    case 'o':					// o = toggle display option (date, hostname, ...)
      get_what_to_toggle();
      redraw_screen();
      break;

    case 'm':					// m = toggle mark (reverse) on the bottom line
      cur_bottom_ll->is_marked ^= 1;
      cur_fmt_serial++;				// (necessary to force redraw)
      redraw_screen();
      break;

    case 'M':						// M = clear all marks
      for (ll = cur_bottom_ll; ll->prev; ll = ll->prev)	
        ll->is_marked = 0;
      cur_fmt_serial++;					// (necessary to force redraw)
      redraw_screen();
      break;

    case '[':						// [ = goto previous marked line
      if (cur_bottom_ll->prev) {
        ll = cur_bottom_ll->prev;
	ll_walk_back(ll, logspad_nlines / 2); 		// begin searching from middle of screen

        while (ll->prev && ! ll->is_marked)
          ll = ll->prev;				// look back for a marked line

        if (ll->is_marked) {				// found !
	  ll_walk_forward(ll, logspad_nlines / 2);	// scroll up 1/2 screen for this line
							// to show up in the middle
          cur_bottom_ll = ll;
          draw_loglines_pad();
          refresh_screen();
        } else
	  transient_message("no mark found backwards");
      }
      break;

    case ']':						// ] = goto next marked line
      if (cur_bottom_ll->next) {
        ll = cur_bottom_ll->next;
	ll_walk_back(ll, logspad_nlines / 2);		// begin searching from middle of screen

        while (ll->next && ! ll->is_marked)
          ll = ll->next;				// look forward for a marked line

        if (ll->is_marked) {				// found !
	  ll_walk_forward(ll, logspad_nlines / 2);	// scroll up 1/2 screen for this line
							// to show up in the middle
          cur_bottom_ll = ll;
          draw_loglines_pad();
          refresh_screen();
        } else
	  transient_message("no mark found forward");
      } else
        transient_message("no mark found forward");
      break;

    case '<':						// < = goto previous day start
      if (cur_bottom_ll->prev) {
        ll = cur_bottom_ll->prev;
	ll_walk_back(ll, logspad_nlines / 2 + 1); 	// begin searching from middle of screen

	while (ll->next && ! ll->day)			// look for a real log line
	  ll = ll->next;				// (with a month and day)
	cur_day = ll->day;
        strncpy(cur_month, ll->month, 3);

        while (ll->prev && ll->day == cur_day && ! strncmp(ll->month, cur_month, 3))
          ll = ll->prev;				// look back for a line with a different date

	ll_walk_forward(ll, logspad_nlines / 2 + 1);	// scroll up 1/2 screen for this line
							// to show up in the middle

        i = logspad_nlines;				// go back 1 screen, then forw 1 screen
        while (ll->prev && i-- > 0)			// to be sure the screen is full of lines
          ll = ll->prev;
        i_am_at_top = (i >= 0);				// position "Top" indicator
	ll_walk_forward(ll, logspad_nlines - 1);
        cur_bottom_ll = ll;

        draw_loglines_pad();
        draw_status_window();
        refresh_screen();
      }
      break;


    case '>':						// > = goto next day start
      if (cur_bottom_ll->next) {
        ll = cur_bottom_ll->next;
	ll_walk_back(ll, logspad_nlines / 2 - 1);	// begin searching from middle of screen

	while (ll->next && ! ll->day)			// look for a real log line
	  ll = ll->next;				// (with a month and day)
	cur_day = ll->day;
        strncpy(cur_month, ll->month, 3);

        while (ll->next && ll->day == cur_day && ! strncmp(ll->month, cur_month, 3))
          ll = ll->next;				// look forward for a line with a different date

	if (ll->day != cur_day || strncmp(ll->month, cur_month, 3)) {	// if found, 
	  ll_walk_forward(ll, logspad_nlines / 2 - 1);	// scroll up 1/2 screen for this line
          cur_bottom_ll = ll;				// to show up in the middle
	} else
	  cur_bottom_ll = tail_line;			// if not found, go to bottom

	i_am_at_top = 0;
        draw_loglines_pad();
        draw_status_window();
        refresh_screen();
      }
      break;

    case 0x1b:						// ESC = stop reverse thread
      if (! reverse_thread_must_stop)			//       or stop running search
        reverse_thread_must_stop = 1;
      else
        kill_search_thread();
      break;

    case '?':						// ? = search backward, give a regexp
    case '/':						// / = search forward,  give a regexp
      if (search_status & SRCH_IN_PROGRESS)
        kill_search_thread();				// kill any search in progress

      get_search_regexp(c == '?');

      if (search_regexp) {				// start search thread
        search_status = SRCH_IN_PROGRESS | (c == '?' ? SRCH_BACKWARDS : 0);
        ret = pthread_create(&search_thread_id , NULL,
		(void *) search_thread, (void *) c);
        if (ret) {
          transient_message("couldn't start search thread (return code %d)", ret);
	  break;
        }

        while ((search_status & SRCH_IN_PROGRESS) && !(search_status & SRCH_FOUND_ONE)) {
	  napms(251);
	  redraw_status_window();

	  if (getch() == 0x1b) {			// ESC
	    kill_search_thread();
	    return;
	  }
        }

        if (search_status & SRCH_FOUND_ONE) {
	  cur_bottom_ll = center_logline_in_screen(search_result_ll);
	  redraw_screen();
        } else
	  transient_message("regexp not found");
      }

      break;

    case 'n':						// n = search next occurence
    case 'N':						// N = search previous occurence
      if (search_regexp) {				// (in same direction as last search)

	backwards = (c == 'N' ? 1 : 0);
	ll = cur_bottom_ll;
	ll_walk_back(ll, logspad_nlines / 2);
	start_ll = ll;

	while (1) {
          redraw_status_window();
	  if (backwards)
	    ll = (ll->prev ? ll->prev : tail_line);
	  else
	    ll = (ll->next ? ll->next : head_line);

	  if (ll->srch_serial != cur_srch_serial)	// if till search thread didn't examine
	    search_logline(ll);				// this line (?), do it

	  if (ll->matched)
	    break;

	  if (ll == start_ll)
	    break;
	}

	if (ll->matched)				// REGEXP FOUND !
	  cur_bottom_ll = center_logline_in_screen(ll);
	else
	  transient_message("regexp not found");

      } else if (c == 'n' || c == 'N')
        transient_message("no regexp entered (type / or ?)");

      redraw_screen();
      break;

    case 's':						// 's' = switch to other logfile
      load_other_logfile(0);				//       (give logfile key)
      break;

    case 'S':						// 'S' = switch to other logfile
      load_other_logfile(1);				//       (give full pathname)
      break;

    case '*':						// '*' = help screen
      help_screen();
      break;

    case '&':						// '&' = display host:path/to/logfile
      transient_message("showing %s:%s (%s)", local_host_name, logfile,
        (mode_ipmon
	 ? "ipfilter logfile"
	 : (mode_httpd_access
	    ? "httpd access logfile"
	    : (mode_httpd_errors
	       ? "httpd errors logfile"
	       : "syslogd logfile"))));
      break;

    case '!':						// '!' = spawn a subshell
      mvcur(0, screen_maxx - 1, screen_maxy - 1, 0);
      nocbreak();
      flushinp();
      endwin();
      fprintf(stderr, "\nvlog: spawning a subshell - type exit to resume\n");
      if (getenv("SHELL"))
        system(getenv("SHELL"));
      else
        system("/bin/sh");
      init_curses();
      redraw_screen();
      break;

    case 'Q':						// 'Q' = quit
      clean_curses_and_exit(0);

    default:
      // fprintf(stderr, "-- KEY = 0x%x --", (int) c);
    }
  }

  i = current_second();
  if (last_typed_key_second + AUTO_REDRAW_DELAY < i ) {
    endwin();						// window automatic redraw
    init_curses();					// every AUTO_REDRAW_DELAY seconds
    redraw_screen();					// when no key are typed
    last_typed_key_second = i;				// (to clean messages garbage)
  }

}


