/*
 * Hash Table Data Type
 * Copyright (C) 1997 Kaz Kylheku <kaz@cafe.net>
 *
 * Free Software License:
 *
 * All rights are reserved by the author, with the following exceptions:
 * Permission is granted to freely reproduce and distribute this software,
 * possibly in exchange for a fee, provided that this copyright notice appears
 * intact. Permission is also granted to adapt this software to produce
 * derivative works, as long as the modified versions carry this copyright
 * notice and additional notices stating that the work has been modified.
 * The copyright extends to translations of this work into other languages,
 * including machine languages. 
 *
 * $Id: hash.c,v 1.12 1997/09/14 17:32:40 kaz Exp $
 * $Name: release_1_4 $
 */

#include <stdlib.h>
#include <stddef.h>
#include <assert.h>
#include "smalloc.h"
#define HASH_IMPLEMENTATION
#include "hash.h"

/*
static char rcsid[] = "$Id: hash.c,v 1.12 1997/09/14 17:32:40 kaz Exp $";
static char right[] = "Copyright (C) 1997 Kaz Kylheku";
*/

#define INIT_SIZE	64	/* must be power of two		*/
#define INIT_SHIFT	6	/* must be log_2(INIT_SIZE)	*/

/*
 * Magic codes that allow safe malloc routines to perform
 * run time type checks.
 */

#ifndef NDEBUG
static const safe_magic_t hash_magic =  "HSHBLOK";
static const safe_magic_t hnode_magic = "HSHNODE";
static const safe_magic_t chain_magic = "HSHCHAN";
#endif

int hash_val_t_bit;

/*
 * Compute the number of bits in the hash_val_t type.  We know that hash_val_t
 * is an unsigned integral type. Thus the highest value it can hold is a
 * Mersenne number (power of two, less one). We initialize a hash_val_t
 * object with this value and then shift bits out one by one while counting.
 * Notes:
 * 1. HASH_VAL_T_MAX is a Mersenne number---one that is one less than a power
 *    of two. This means that its binary representation consists of all one
 *    bits, and hence ``val'' is initialized to all one bits.
 * 2. We reset the bit count to zero in case this function is invoked more than
 *    once.
 * 3. While bits remain in val, we increment the bit count and shift it to the
 *    right, replacing the topmost bit by zero.
 */

static void compute_bits(void)
{
    hash_val_t val = HASH_VAL_T_MAX;	/* 1 */

    hash_val_t_bit = 0;			/* 2 */
    while (val) {	/* 3 */
	hash_val_t_bit++;
	val >>= 1;
    }
}

/*
 * Verify whether the given argument is a power of two.
 */

static int is_power_of_two(hash_val_t arg)
{
    if (arg == 0)
	return 0;
    while ((arg & 1) == 0)
	arg >>= 1;
    return (arg == 1);
}

/*
 * Compute a shift amount from a given table size 
 */

static int compute_shift(hash_val_t size)
{
    int shift = hash_val_t_bit;

    assert (is_power_of_two(size));
    assert (hash_val_t_bit != 0);

    size >>= 1;
    while (size) {
	shift--;
	size >>= 1;
    }

    assert (hash_val_t_bit >> shift == size - 1);

    return shift;
}

/*
 * Initialize the table of pointers to null.
 */

static void clear_table(hash_t *hash)
{
    hash_val_t i;

    for (i = 0; i < hash->nchains; i++)
	hash->table[i] = NULL;
}

/*
 * Double the size of a dynamic table. This works as follows. Each chain splits
 * into two adjacent chains.  The shift amount increases by one, exposing an
 * additional bit of each hashed key. For each node in the original chain, the
 * value of this newly exposed bit will decide which of the two new chains will
 * receive the node: if the bit is 1, the chain with the higher index will have
 * the node, otherwise the lower chain will receive the node. In this manner,
 * the hash table will continue to function exactly as before without having to
 * rehash any of the keys.
 * Notes:
 * 1.  Overflow check.
 * 2.  The new number of chains is twice the old number of chains.
 * 3.  The new shift value is one less than the prior shift value, revealing a
 *     new bit in all hashed keys.
 * 4.  Allocate a new table of chain pointers that is twice as large as the
 *     previous one.
 * 5.  If the reallocation was successful, we perform the rest of the growth
 *     algorithm, otherwise we do nothing.
 * 6.  The exposed_bit variable holds a mask with which each hashed key can be
 *     AND-ed to test the value of its newly exposed bit.
 * 7.  A backward loop over all the chain pairs. The chain variable indexes
 *     the lower of each adjacent chain pair in turn. The oldchain variable
 *     loops over the lower half of the table which contains all of the 
 *     chain pointers from before the growth which must be spaced apart.
 * 8.  Each chain from the original table is copied to the lower member of each
 *     new chain pair. The upper member is set to a null pointer (empty chain).
 *     The result is a table in which even numbered chain has zero or more
 *     nodes inherited from the corresponding chain in the smaller table, and
 *     each odd chain is empty.
 * 9.  Since we have relocated the table of pointers, we have to fix the
 *     back-reference from the first node of each non-empty chain so it
 *     properly refers to the moved pointer.
 * 10. We loop over the even chain looking for any nodes whose exposed bit is
 *     set. Such nodes are removed from the even chain and put into its odd
 *     sister.
 * 11. Before moving the node to the other chain, we remember what the next
 *     node is so we can coninue the loop. We have to do this because we will
 *     be overwriting the node's next pointer when we insert it to its new
 *     home.
 * 12. The next node's back pointer must be updated to skip to the previous
 *     node.
 * 13. The deleted node's back pointer must be updated to refer to the next
 *     node.
 * 14. We insert the node at the beginning of the odd chain.
 * 15. If the chain where we just inserted the node was not empty, we must
 *     update the back-pointer of what used to be the first node to correctly
 *     point to the next pointer in the new node.
 * 16. We have finished dealing with the chains and nodes. We now update
 *     the various bookeeping fields of the hash structure.
 */

static void grow_table(hash_t *hash)
{
    hash_val_t chain, oldchain, nchains, exposed_bit;
    hnode_t **newtable, *hptr;
    int shift;

    assert (2 * hash->nchains > hash->nchains);	/* 1 */

    nchains = 2 * hash->nchains;	/* 2 */
    shift = hash->shift - 1;		/* 3 */

    newtable = safe_realloc(hash->table, sizeof *newtable * hash->nchains,
    	sizeof *newtable * nchains, &chain_magic);	/* 4 */

    if (newtable) {	/* 5 */
	exposed_bit = 1 << shift;	/* 6 */
	for (oldchain = nchains/2 - 1, chain = nchains-2;
		chain <= nchains-2; oldchain--, chain -= 2) {	/* 7 */
	    hptr = newtable[chain] = newtable[oldchain];	/* 8 */
	    if (hptr)	/* 9 */
		hptr->pself = newtable + chain;
	    newtable[chain + 1] = NULL;
	    while (hptr) {	/* 10 */
		if ((hptr->hkey & exposed_bit)) {
		    hnode_t *next = hptr->next;		/* 11 */
		    /* delete node in one chain */
		    if (next)				/* 12 */
			next->pself = hptr->pself;
		    *hptr->pself = next;		/* 13 */
		    /* prepend node to other chain */
		    hptr->next = newtable[chain+1];	/* 14 */
		    hptr->pself = &newtable[chain+1];
		    newtable[chain+1] = hptr;
		    if (hptr->next)			/* 15 */
			hptr->next->pself = &hptr->next;
		    hptr = next;
		} else {
		    hptr = hptr->next;
		}
	    }
	}
	hash->table = newtable;		/* 16 */
	hash->shift = shift;
	hash->nchains = nchains;
	hash->lowmark *= 2;
	hash->highmark *= 2;
    }
    assert (hash_verify(hash));
}

/*
 * Cut a table size in half. This is done by folding together adjacent chains
 * and populating the lower half of the table with these chains. The chains are
 * simply spliced together. Once this is done, the whole table is reallocated
 * to a smaller object.
 * Notes:
 * 1.  It is illegal to have a hash table with one slot. This would mean that
 *     hash->shift is equal to hash_val_t_bit, an illegal shift value.
 *     Also, other things could go wrong, such as hash->lowmark becoming zero.
 * 2.  Looping over each adjacent chain of pairs, the lo_chain is set to
 *     reference the lower-numbered member of the pair, whereas hi_chain
 *     is the index of the higher-numbered one.
 * 3.  The intent here is to compute a pointer to the last node of the
 *     lower chain into the lo_tail variable. If this chain is empty,
 *     lo_tail ends up with a null value.
 * 4.  If the lower chain is not empty, we have to merge chains, but only
 *     if the upper chain is also not empty. In either case, the lower chain
 *     will come first, with the upper one appended to it.
 * 5.  The first part of the join is done by having the tail node of the lower
 *     chain point to the head node of the upper chain. If the upper chain
 *     is empty, this is remains a null pointer.
 * 6.  If the upper chain is non-empty, we must do the additional house-keeping
 *     task of ensuring that the upper chain's first node's back-pointer
 *     references the tail node of the lower chain properly.
 * 7.  All that's left is to update the the table entry to point to the
 *     first node in what used to be the low chain.
 * 8.  If the low chain is empty, but the high chain is not, then the
 *     high chain simply becomes the new chain.
 * 9.  Otherwise if both chains are empty, then the merged chain is also empty.
 * 10. All the chain pointers are in the lower half of the table now, so
 *     we reallocate it to a smaller object. This, of course, invalidates
 *     all pointer-to-pointers which reference into the table from the
 *     first node of each chain.
 * 11. Though it's unlikely, the reallocation may fail. In this case we
 *     pretend that the table _was_ reallocated to a smaller object.
 * 12. This loop performs the housekeeping task of updating the back pointers
 *     from the first node of each chain so that they reference their 
 *     corresponding table entries.
 * 13. Finally, update the various table parameters to reflect the new size.
 */

static void shrink_table(hash_t *hash)
{
    hash_val_t chain, oldchain, nchains;
    hnode_t **newtable, *lo_tail, *lo_chain, *hi_chain;

    assert (hash->nchains >= 2);			/* 1 */
    nchains = hash->nchains / 2;

    for (chain = 0, oldchain = 0; chain < nchains; chain++, oldchain+=2) {
	lo_chain = hash->table[oldchain];		/* 2 */
	hi_chain = hash->table[oldchain + 1];
	for (lo_tail=lo_chain; lo_tail && lo_tail->next; lo_tail=lo_tail->next)
	    ;	/* 3 */
	if (lo_chain) {					/* 4 */
	    lo_tail->next = hi_chain;			/* 5 */
	    if (hi_chain)				/* 6 */
		hi_chain->pself = &lo_tail->next;
	    hash->table[chain] = lo_chain;		/* 7 */
	} else if (hi_chain) {				/* 8 */
	    hash->table[chain] = hi_chain;
	} else {
	    hash->table[chain] = NULL;			/* 9 */
	}
    }
    newtable = safe_realloc(hash->table, sizeof *newtable * hash->nchains,
	sizeof *newtable * nchains, &chain_magic);	/* 10 */
    if (newtable)					/* 11 */
	hash->table = newtable;
    for (chain = 0; chain < nchains; chain++)		/* 12 */
	if (hash->table[chain])
	    hash->table[chain]->pself = &hash->table[chain];
    hash->shift += 1;			/* 13 */
    hash->nchains = nchains;
    hash->lowmark /= 2;
    hash->highmark /= 2;
    assert (hash_verify(hash));
}


/*
 * Create a dynamic hash table. Both the hash table structure and the table
 * itself are dynamically allocated. Furthermore, the table is extendible in
 * that it will automatically grow as its load factor increases beyond a
 * certain threshold.
 * Notes:
 * 1. If the number of bits in the hash_val_t type has not been computed yet,
 *    we do so here, because this is likely to be the first function that the
 *    user calls.
 * 2. Safe malloc is used for added checking. The checking is disabled by
 *    defining NDEBUG, which turns the safe malloc routines (via the
 *    preprocessor) into direct calls to malloc, free, etc.
 * 3. If a hash table control structure is successfully allocated, we
 *    proceed to initialize it. Otherwise we return a null pointer.
 * 4. Using the safe allocator, we try to allocate the table of hash
 *    chains.
 * 5. If we were able to allocate the hash chain table, we can finish
 *    initializing the hash structure and the table. Otherwise, we must
 *    backtrack by freeing the hash structure.
 * 6. INIT_SIZE should be a power of two. The high and low marks are always set
 *    to be twice the table size and half the table size respectively. When the
 *    number of nodes in the table grows beyond the high size (beyond load
 *    factor 2), it will double in size to cut the load factor down to about
 *    about 1. If the table shrinks down to or beneath load factor 0.5,
 *    it will shrink, bringing the load up to about 1. However, the table
 *    will never shrink beneath INIT_SIZE even if it's emptied.
 * 7. This indicates that the table is dynamically allocated and dynamically
 *    resized on the fly. A table that has this value set to zero is
 *    assumed to be statically allocated and will not be resized.
 * 8. The table of chains must be properly reset to all null pointers.
 */

hash_t *hash_create(hashcount_t maxcount, hash_comp_t compfun,
	hash_fun_t hashfun)
{
    hash_t *hash;

    if (hash_val_t_bit == 0)	/* 1 */
	compute_bits();

    hash = safe_malloc(sizeof *hash, &hash_magic);	/* 2 */

    if (hash) {		/* 3 */
	hash->table = safe_malloc(sizeof *hash->table * INIT_SIZE,
		&chain_magic);	/* 4 */
	if (hash->table) {	/* 5 */
	    hash->nchains = INIT_SIZE;		/* 6 */
	    hash->highmark = INIT_SIZE * 2;
	    hash->lowmark = INIT_SIZE / 2;
	    hash->count = 0;
	    hash->maxcount = maxcount;
	    hash->compare = compfun;
	    hash->hash = hashfun;
	    hash->shift = hash_val_t_bit - INIT_SHIFT;
	    hash->dynamic = 1;			/* 7 */
	    clear_table(hash);			/* 8 */
	    assert (hash_verify(hash));
	    return hash;
	} 
	safe_free(hash, sizeof *hash, &hash_magic);
    }

    return NULL;
}

/*
 * Free a dynamic hash table structure.
 */

void hash_destroy(hash_t *hash)
{
    assert (hash_val_t_bit != 0);
    assert (hash_isempty(hash));
    safe_free(hash->table, sizeof *hash->table * hash->nchains, &chain_magic);
    safe_free(hash, sizeof *hash, &hash_magic);
}

/*
 * Initialize a user supplied hash structure. The user also supplies a table of
 * chains which is assigned to the hash structure. The table is static---it
 * will not grow or shrink.
 * 1. See note 1. in hash_create().
 * 2. The user supplied array of pointers hopefully contains nchains nodes.
 * 3. See note 7. in hash_create().
 * 4. We must dynamically compute the shift from the given power of two table
 *    size. The shift is equal to    hash_val_t_bit - log_2(nchains).
 * 5. The user supplied table can't be assumed to contain null pointers,
 *    so we reset it here.
 */

hash_t *hash_init(hash_t *hash, hashcount_t maxcount,
	hash_comp_t compfun, hash_fun_t hashfun, hnode_t **table,
	hashcount_t nchains)
{
    if (hash_val_t_bit == 0)	/* 1 */
	compute_bits();

    assert (is_power_of_two(nchains));

    hash->table = table;	/* 2 */
    hash->nchains = nchains;
    hash->count = 0;
    hash->maxcount = maxcount;
    hash->compare = compfun;
    hash->hash = hashfun;
    hash->dynamic = 0;		/* 3 */
    hash->shift = compute_shift(nchains);	/* 4 */
    clear_table(hash);		/* 5 */

    assert (hash_verify(hash));

    return hash;
}

/*
 * Reset the hash scanner so that the next element retrieved by
 * hash_scan_next() shall be the first element on the first non-empty chain. 
 * Notes:
 * 1. Locate the first non empty chain.
 * 2. If an empty chain is found, remember which one it is and set the next
 *    pointer to refer to its first element.
 * 3. Otherwise if a chain is not found, set the next pointer to NULL
 *    so that hash_scan_next() shall indicate failure.
 */

void hash_scan_begin(hscan_t *scan, hash_t *hash)
{
    hash_val_t nchains = hash->nchains;
    hash_val_t chain;

    scan->hash = hash;

    /* 1 */

    for (chain = 0; chain < nchains && hash->table[chain] == 0; chain++)
	;

    if (chain < nchains) {	/* 2 */
	scan->chain = chain;
	scan->next = hash->table[chain];
    } else {			/* 3 */
	scan->next = NULL;
    }
}

/*
 * Retrieve the next node from the hash table, and update the pointer
 * for the next invocation of hash_scan_next(). 
 * Notes:
 * 1. Remember the next pointer in a temporary value so that it can be
 *    returned.
 * 2. This assertion essentially checks whether the module has been properly
 *    initialized. The first point of interaction with the module should be
 *    either hash_create() or hash_init(), both of which set hash_val_t_bit to
 *    a non zero value.
 * 3. If the next pointer we are returning is not NULL, then the user is
 *    allowed to call hash_scan_next() again. We prepare the new next pointer
 *    for that call right now. That way the user is allowed to delete the node
 *    we are about to return, since we will no longer be needing it to locate
 *    the next node.
 * 4. If there is a next node in the chain (next->next), then that becomes the
 *    new next node, otherwise ...
 * 5. We have exhausted the current chain, and must locate the next subsequent
 *    non-empty chain in the table.
 * 6. If a non-empty chain is found, the first element of that chain becomes
 *    the new next node. Otherwise there is no new next node and we set the
 *    pointer to NULL so that the next time hash_scan_next() is called, a null
 *    pointer shall be immediately returned.
 */


hnode_t *hash_scan_next(hscan_t *scan)
{
    hnode_t *next = scan->next;		/* 1 */
    hash_t *hash = scan->hash;
    hash_val_t chain = scan->chain + 1;
    hash_val_t nchains = hash->nchains;

    assert (hash_val_t_bit != 0);	/* 2 */

    if (next) {			/* 3 */
	if (next->next) {	/* 4 */
	    scan->next = next->next;
	} else {
	    while (chain < nchains && hash->table[chain] == 0)	/* 5 */
	    	chain++;
	    if (chain < nchains) {	/* 6 */
		scan->chain = chain;
		scan->next = hash->table[chain];
	    } else {
		scan->next = NULL;
	    }
	}
    }
    return next;
}

/*
 * Insert a node into the hash table.
 * Notes:
 * 1. It's illegal to insert more than the maximum number of nodes. The client
 *    should verify that the hash table is not full before attempting an
 *    insertion.
 * 2. The same key may not be inserted into a table twice.
 * 3. Ensure that the shift amount is within defined boundaries.
 * 4. If the table is dynamic and the load factor is already at >= 2,
 *    grow the table.
 * 5. We take the top N bits of the hash value to derive the chain index,
 *    where N is the base 2 logarithm of the size of the hash table. 
 */

void hash_insert(hash_t *hash, hnode_t *node, void *key)
{
    hash_val_t hkey, chain;

    assert (hash_val_t_bit != 0);
    assert (hash->count < hash->maxcount);	/* 1 */
    assert (hash_lookup(hash, key) == NULL);	/* 2 */
    assert (hash->shift < hash_val_t_bit);	/* 3 */

    if (hash->dynamic && hash->count >= hash->highmark)	/* 4 */
	grow_table(hash);

    hkey = hash->hash(key);
    chain = hkey >> hash->shift;	/* 5 */

    node->key = key;
    node->hkey = hkey;
    node->pself = hash->table + chain;
    node->next = hash->table[chain];
    if (node->next)
	node->next->pself = &node->next;
    hash->table[chain] = node;
    hash->count++;

    assert (hash_verify(hash));
}

/*
 * Find a node in the hash table and return a pointer to it.
 * Notes:
 * 1. We hash the key and keep the entire hash value. As an optimiation, when
 *    we descend down the chain, we can compare hash values first and only if
 *    hash values match do we perform a full key comparison. 
 * 2. To locate the chain from among 2^N chains, we look at the top N bits of
 *    the hash value.
 * 3. Looping through the chain, we compare the stored hash value inside each
 *    node against our computed hash. If they match, then we do a full
 *    comparison between the unhashed keys. If these match, we have located the
 *    entry.
 */

hnode_t *hash_lookup(hash_t *hash, void *key)
{
    hash_val_t hkey, chain;
    hnode_t *nptr;

    assert (hash->shift < hash_val_t_bit);
    hkey = hash->hash(key);		/* 1 */
    chain = hkey >> hash->shift;	/* 2 */

    for (nptr = hash->table[chain]; nptr; nptr = nptr->next) {	/* 3 */
	if (nptr->hkey == hkey && hash->compare(nptr->key, key) == 0)
	    return nptr;
    }

    return NULL;
}

/*
 * Delete the given node from the hash table. This is easy, because each node
 * contains a back pointer to the previous node's next pointer.
 * Notes:
 * 1. The node must belong to this hash table, and its key must not have
 *    been tampered with.
 * 2. If there is a next node, then we must update its back pointer to
 *    skip this node.
 * 3. We must update the pointer that is pointed at by the back-pointer
 *    to skip over the node that is being deleted and instead point to
 *    the successor (or to NULL if the node being deleted is the last one).
 */

hnode_t *hash_delete(hash_t *hash, hnode_t *node)
{
    assert (hash_lookup(hash, node->key) == node);	/* 1 */
    assert (hash_val_t_bit != 0);

    if (hash->dynamic && hash->count <= hash->lowmark
	    && hash->count > INIT_SIZE)
	shrink_table(hash);

    if (node->next)	/* 2 */
	node->next->pself = node->pself;
    *node->pself = node->next;	/* 3 */
    hash->count--;
    assert (hash_verify(hash));
    return node;
}

/*
 *  Exactly like hash_delete, except does not trigger table shrinkage. This is to be
 *  used from within a hash table scan operation. See notes for hash_delete.
 */

hnode_t *hash_scan_delete(hash_t *hash, hnode_t *node)
{
    assert (hash_lookup(hash, node->key) == node);	/* 1 */
    assert (hash_val_t_bit != 0);

    if (node->next)	/* 2 */
	node->next->pself = node->pself;
    *node->pself = node->next;	/* 3 */
    hash->count--;
    assert (hash_verify(hash));
    return node;
}

/*
 * Verify whether the given object is a valid hash table. This means
 * Notes:
 * 1. If the hash table is dynamic, verify whether the high and
 *    low expansion/shrinkage thresholds are powers of two.
 * 2. The shift count must be non-negative and less than the number of
 *    bits in hash_val_t at all times.
 * 3. For each chain, verify whether the back pointers are correctly
 *    set. Count all nodes in the table, and test each hash value
 *    to see whether it is correct for the node's chain.
 */

int hash_verify(hash_t *hash)
{
    hashcount_t count = 0;
    hash_val_t chain;
    hnode_t **npp;

    if (hash->dynamic) {	/* 1 */
	if (hash->lowmark >= hash->highmark)
	    return 0;
	if (!is_power_of_two(hash->highmark))
	    return 0;
	if (!is_power_of_two(hash->lowmark))
	    return 0;
    }

    if (hash->shift < 0 || hash->shift >= hash_val_t_bit)	/* 2 */
	return 0;

    for (chain = 0; chain < hash->nchains; chain++) {	/* 3 */
	for (npp = hash->table + chain; *npp; npp = &(*npp)->next) {
	    if ((*npp)->pself != npp)
		return 0;
	    if ((*npp)->hkey >> hash->shift != chain)
		return 0;
	    count++;
	}
    }

    if (count != hash->count)
	return 0;

    return 1;
}

/*
 * Test whether the hash table is full and return 1 if this is true,
 * 0 if it is false.
 */

#undef hash_isfull
int hash_isfull(hash_t *hash)
{
    return hash->count == hash->maxcount;
}

/*
 * Test whether the hash table is empty and return 1 if this is true,
 * 0 if it is false.
 */

#undef hash_isempty
int hash_isempty(hash_t *hash)
{
    return hash->count == 0;
}

/*
 * Create a hash table node dynamically and assign it the given data.
 */

hnode_t *hnode_create(void *data)
{
    hnode_t *node = safe_malloc(sizeof *node, &hnode_magic);
    if (node) {
	node->data = data;
	node->next = NULL;
	node->pself = NULL;
    }
    return node;
}

/*
 * Initialize a client-supplied node 
 */

hnode_t *hnode_init(hnode_t *hnode, void *data)
{
    hnode->data = data;
    hnode->next = NULL;
    hnode->pself = NULL;
    return hnode;
}

/*
 * Destroy a dynamically allocated node.
 */

void hnode_destroy(hnode_t *hnode)
{
    safe_free(hnode, sizeof *hnode, &hnode_magic);
}

#undef hnode_put
void hnode_put(hnode_t *node, void *data)
{
    node->data = data;
}

#undef hnode_get
void *hnode_get(hnode_t *node)
{
    return node->data;
}

#undef hnode_getkey
void *hnode_getkey(hnode_t *node)
{
    return node->key;
}

#undef hash_count
hashcount_t hash_count(hash_t *hash)
{
    return hash->count;
}

#undef hash_size
hashcount_t hash_size(hash_t *hash)
{
    return hash->nchains;
}


#ifdef TEST_MAIN

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>

typedef char input_t[256];

static int tokenize(char *string, ...)
{
    char **tokptr; 
    va_list arglist;
    int tokcount = 0;

    va_start(arglist, string);
    tokptr = va_arg(arglist, char **);
    while (tokptr) {
	while (*string && isspace(*string))
	    string++;
	if (!*string)
	    break;
	*tokptr = string;
	while (*string && !isspace(*string))
	    string++;
	tokptr = va_arg(arglist, char **);
	tokcount++;
	if (!*string)
	    break;
	*string++ = 0;
    }
    va_end(arglist);

    return tokcount;
}

static int comparef(const void *key1, const void *key2)
{
    return strcmp(key1, key2);
}

static char *dupstring(char *str)
{
    int sz = strlen(str) + 1;
    char *new = malloc(sz);
    memcpy(new, str, sz);
    return new;
}

static hash_val_t hashstr(const void *key)
{
    const char *str = key;
    hash_val_t acc = 0;

    while (*str) {
	acc ^= *str++;
	acc = (acc >> 4) | ((acc & 0xf) << (hash_val_t_bit - 4));
    }

    return acc;
}

int main(void)
{
    input_t in;
    hash_t *h = hash_create(HASHCOUNT_T_MAX, comparef, hashstr);
    hnode_t *hn;
    hscan_t hs;
    char *tok1, *tok2, *key, *val;
    int prompt = 0;

    char *help =
	"a <key> <val>          add value to hash table\n"
	"d <key>                delete value from hash table\n"
	"l <key>                lookup value in hash table\n"
	"n                      show size of hash table\n"
	"c                      show number of entries\n"
	"t                      dump whole hash table\n"
	"+                      increase hash table (private func)\n"
	"-                      decrease hash table (private func)\n"
	"b                      print hash_t_bit value\n"
	"p                      turn prompt on\n"
	"q                      quit";

    if (!h)
	puts("hash_create failed");

    for (;;) {
	if (prompt)
	    putchar('>');
	fflush(stdout);

	if (!fgets(in, sizeof(input_t), stdin))
	    break;

	switch(in[0]) {
	    case '?':
		puts(help);
		break;
	    case 'b':
		printf("%d\n", hash_val_t_bit);
		break;
	    case 'a':
		if (tokenize(in+1, &tok1, &tok2, (char **) 0) != 2) {
		    puts("what?");
		    break;
		}
		key = dupstring(tok1);
		val = dupstring(tok2);
		hn = hnode_create(val);
		if (!hn) {
		    puts("hnode_create failed");
		    free(key);
		    free(val);
		    break;
		}
		hash_insert(h, hn, key);
		break;
	    case 'd':
		if (tokenize(in+1, &tok1, (char **) 0) != 1) {
		    puts("what?");
		    break;
		}
		hn = hash_lookup(h, tok1);
		if (!hn) {
		    puts("hash_lookup failed");
		    break;
		}
		key = hnode_get(hn);
		val = hnode_getkey(hn);
		hash_delete(h, hn);
		hnode_destroy(hn);
		free(key);
		free(val);
		break;
	    case 'l':
		if (tokenize(in+1, &tok1, (char **) 0) != 1) {
		    puts("what?");
		    break;
		}
		hn = hash_lookup(h, tok1);
		if (!hn) {
		    puts("hash_lookup failed");
		    break;
		}
		val = hnode_get(hn);
		puts(val);
		break;
	    case 'n':
		printf("%lu\n", (unsigned long) hash_size(h));
		break;
	    case 'c':
		printf("%lu\n", (unsigned long) hash_count(h));
		break;
	    case 't':
		hash_scan_begin(&hs, h);
		while ((hn = hash_scan_next(&hs)))
		    printf("%s\t%s\n", (char*) hnode_getkey(hn),
			    (char*) hnode_get(hn));
		break;
	    case '+':
		grow_table(h);		/* private function	*/
		break;
	    case '-':
		shrink_table(h);	/* private function	*/
		break;
	    case 'q':
		exit(0);
		break;
	    case '\0':
		break;
	    case 'p':
		prompt = 1;
		break;
	    default:
		putchar('?');
		putchar('\n');
		break;
	}
    }

    return 0;
}

#endif
