/*
 * ipfmodule.c
 *
 * Curses based IP-Filter logs viewer
 *
 * Copyright (C) 2001  Pierre Berthomier
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2
 * as published by the Free Software Foundation, provided that this notice
 * is preserved in its entirety and due credit is given to the original
 * author and the contributors.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#include "vlog.h"

// -----------------------------------------------------------------------------

char     ipf_conf_dir[SMALLSTR];
char     ipf_conf_name[SMALLSTR];
char     nat_conf_name[SMALLSTR];
hash_t * h_tcp_service_name_to_port;
hash_t * h_udp_service_name_to_port;
hash_t * h_ipf_conf_to_isba_rule;
hash_t * h_group_num_to_isba_rule;

char *   icmp_type_code[19][16];
char *   ipproto_name[PROTO_NAME_MAX];
char *   tcp_serv_name[SERV_NAME_MAX];
char *   udp_serv_name[SERV_NAME_MAX];

// ------------------------------------------------------------------------------

/* hash tables: key comparison and hash functions for ascii keys */

static int ascii_comparef(const void * key1, const void * key2)
{
  return strcmp(key1, key2);
}

/* hash function for ascii keys */

static hash_val_t ascii_hashstr(const void * key)
{
    const char * str = key;
    hash_val_t acc = 0;

    while (*str) {
	acc ^= *str++;
	acc = (acc >> 4) | ((acc & 0xf) << (hash_val_t_bit - 4));
    }

    return acc;
}

/* hash-tables: key comparison and hash functions for int keys */

static int int32_comparef(const void * key1, const void * key2) {
  // return memcmp(key1, key2, 4);
  return ((int) key1 < (int) key2
	  ? -1
	  : ((int) key1 > (int) key2
	     ? +1
	     : 0));
}

/* hash function for int32 keys */

static hash_val_t int32_hashstr(const void * key)
{

  return (hash_val_t) key;

//register unsigned char * k = (unsigned char *) key;
//return (((int) k[3] << 24) | ((int) k[2] << 16)
//	| ((int) k[1] <<  8) |  (int) k[0]);
}

/* Create and returns a hash table for strings. Title needed for error message*/

hash_t * new_ascii_hash_table(char * title)
{
hash_t * ht;

  if (!(ht = hash_create(HASHCOUNT_T_MAX, ascii_comparef, ascii_hashstr)))
    err("new_ascii_hash_table: hash_create failed for table '%s'", title);
  return(ht);
}

/* Create and returns a hash table for int32s. Title needed for error message */

hash_t * new_int32_hash_table(char * title)
{
hash_t * ht;

    if (!(ht = hash_create(HASHCOUNT_T_MAX, int32_comparef, int32_hashstr)))
      err("new_int32_hash_table: hash_create failed for table '%s'", title);
    return(ht);
}

// ------------------------------------------------------------------------------

/* Guess ipfilter conf files paths. Return TRUE if guessed. */

int guess_ipf_conf_names()
{
  struct stat statbuf;
  char        buf[SMALLSTR];

  sprintf(buf, "/etc/opt/ipf/ipf.conf");
  if (! stat(buf, &statbuf)) {
    sprintf(ipf_conf_dir,  "/etc/opt/ipf");
    sprintf(ipf_conf_name, "ipf.conf");		// /etc/opt/ipf/ipf.conf
    sprintf(nat_conf_name, "ipnat.conf");	// /etc/opt/ipf/ipnat.conf

  } else {
    sprintf(buf, "/etc/ipf.conf");
    if (! stat(buf, &statbuf)) {
      sprintf(ipf_conf_dir,  "/etc");
      sprintf(ipf_conf_name, "ipf.conf");	// /etc/ipf.conf
      sprintf(nat_conf_name, "ipnat.conf");	// /etc/ipnat.conf

    } else {
      sprintf(buf, "/etc/ipf.rules");
      if (! stat(buf, &statbuf)) {
        sprintf(ipf_conf_dir,  "/etc");
        sprintf(ipf_conf_name, "ipf.rules");	// /etc/ipf.rules
        sprintf(nat_conf_name, "ipnat.rules");	// /etc/ipnat.rules
      } else
        return FALSE;
    }
  }

  return TRUE;
}

/* This routine stores the source ipfilter rules into a hash table
 * along with the isba source rule number, extracted from /etc/ipf.conf.
 * The goal is to be able to get isba rulenum, given a @group:rule
 * ipfilter rule spec, so as to display the source rule number.
 *
 * 	  key: ipf rule text
 * 	value: isba rulenum
 */

int ipf_conf_to_hash_table()
{
  FILE *    fd;
  char      ipf_conf[SMALLSTR];
  char      buf[BUFSIZ];
  char *    s;
  int       isbarule;
  hnode_t * hn;
  int       found_an_isbanum = 0;
  int       isba_version_ok  = 0;
  int       line = 0;
  int       isba_major, isba_minor;

  if (guess_ipf_conf_names()) {
    sprintf(ipf_conf, "%s/%s", ipf_conf_dir, ipf_conf_name);
    if ((fd = fopen(ipf_conf, "r"))) {

      h_ipf_conf_to_isba_rule = new_ascii_hash_table("ipf conf to isba rulenums");

      while (fgets(buf, BUFSIZ, fd)) {
        line++;

	if (buf[0] == '#') {				// comments: only look for the 'Isba version : ' line
	  if (!isba_version_ok && (s = strstr(buf, "Isba version : "))) {
	    s = (char *) (s + strlen("Isba version : "));
	    if (2 == sscanf(s, "%d.%d", &isba_major, &isba_minor)) {
	      isba_version_ok = (isba_major > 1) || (isba_major == 1 && isba_minor >= 1);	// isba-1.1+ needed for
	    }											// displaying isba rulenums
	  }

	} else if (buf[0] != '\n') {			// avoid empty lines
	  if ((s = strchr(buf, '#'))) {			// if the first '#' is followed by
	    if (isdigit((int) *++s)) {			// a digit, I assume this is an
	      sscanf(s, "%d", &isbarule);		// isba source rule number.
	      s--; s--;					// get back to the '#' char
	      while (*--s == ' ') ;			// get back, skipping all space chars
	      *++s = '\0';				// terminate string here (end of ipf rule)
	      found_an_isbanum = 1;

	      if (! hash_lookup(h_ipf_conf_to_isba_rule, buf)) {
	      
	        hn = hnode_create((void *) isbarule);
	        if (!hn)
	          err("ipfilter_rules_to_hash_table(): hnode_create failed");
		// fprintf(stderr, "INSERT: isbarule(%d) <-- '%s'\n", isbarule, buf);
	        hash_insert(h_ipf_conf_to_isba_rule, hn, tstrdup(buf));	// rule text = key, isba rule num = data

	      } else {					// the same rule already exists !
	        fprintf(stderr, "*** %s: duplicate rule in line %d !!\n", ipf_conf, line);
	        sleep(2);
	      }

	    }
	  }
	}
      }
      fclose(fd);
      return (found_an_isbanum && isba_version_ok);
    }
  }
  
  return FALSE;
}

/* return the port number associated with the service name in string s,
 * or -1 if s already contains a number
 */

int service_name_to_port(
  int    is_tcp,
  char * s)
{
  char      service[SMALLSTR];
  hnode_t * hn;

  if (isdigit((int) *s))
    return -1;

  copy_to_nth_c(s, service, 1, ' ');
  if (is_tcp
   && (hn = hash_lookup(h_tcp_service_name_to_port, service)))
    return (int) hnode_get(hn);			// do I know this tcp service ?

  else if ((hn = hash_lookup(h_udp_service_name_to_port, service)))
    return (int) hnode_get(hn);			// do I know this udp service ?

  return -1;
}


/* read the rules from kernel (ipfstat -ion)
 * and create the hash table h_group_num_to_isba_rule:
 *
 *   (ipfilter group, ipfilter rulenum) ---> (isba rulenum)
 */

void ipfstat_ion_to_hash_table()
{
  FILE *    fd;
  char      buf[BUFSIZ];
  char      buf2[BUFSIZ];
  char *    rule;
  char *    s;
  int       group, num, group_dir_num;
  hnode_t * hn;
  int       isbanum;
  int       nrules_i_didnt_find = 0;
  int       port;
  int       is_tcp;
  int       nchars1;
  char *    t;

    if (! (fd = popen("ipfstat -ion 2>/dev/null", "r")))
      if (! getuid())							// I'm root and ipfstat
        return;								// fails => no way.
    if (! fd)
      if (! (fd = popen("sudo ipfstat -ion 2>/dev/null", "r")))		// I'm not root: try with sudo
        return;

    h_group_num_to_isba_rule = new_int32_hash_table("group:num to isba rule");

    while (fgets(buf, BUFSIZ, fd)) {
      sscanf(buf, "@%d", &num);
      s = index(buf, '\n');				// remove trailing \n
      *s = '\0';

      rule = buf;
      skip_to_nth_c(rule, 1, ' ');			// rule points to rule (skipped @NN)

      if ((s = strstr(rule, "group ")))
	sscanf(s, "group %d", &group);
      else
        group = 0;

      // now I'll replace any tcp/udp service name present with the port associated
      // (for some ports ipfilter says "port = sunrpc" instead of "port = 111")

      if ((s = strstr(rule, "port = "))) {
        is_tcp = (strstr(rule, "proto tcp") != NULL);
	t = (char *) ((int) s + 7);
	if ((port = service_name_to_port(is_tcp, t)) > 0) {
	  nchars1 = (int) t - (int) rule;
	  skip_to_nth_c(t, 1, ' ');				// replace service name with port number
	  sprintf(buf2, "%*.*s%d %s", nchars1, nchars1, rule, port, t);
	  rule = buf2;
        }

	s++;							// same thing for the destination port
        if ((s = strstr(s, "port = "))) {			// (if any)
	  t = (char *) ((int) s + 7);
	  if ((port = service_name_to_port(is_tcp, t)) > 0) {
	    nchars1 = (int) t - (int) rule;
	    skip_to_nth_c(t, 1, ' ');				// replace service name with port number
	    sprintf(buf, "%*.*s%d %s", nchars1, nchars1, rule, port, t);
	    rule = buf;
          }
	}
      }

      // fprintf(stderr, "RULE = %s.\n", rule);

      if ((hn = hash_lookup(h_ipf_conf_to_isba_rule, rule))) {	// is this rule in my ipf_conf hash table ?
	isbanum = (int) hnode_get(hn);				// yes! get the associated isba rule num
        hn = hnode_create((void *) isbanum);			// and put it in the h_group_num_to_isba_rule
        if (!hn)						// table.
          err("ipfstat_ion_to_hash_table(): hnode_create failed");

	// fprintf(stderr, "@%d:%d --> %d (%s)\n", group, num, isbanum, rule);

	group_dir_num = (int) ((group << 16) + num + (strstr(rule, " out ") ? 32768 : 0));
        hash_insert(h_group_num_to_isba_rule, hn, (void *) group_dir_num);
	  // key  = (group | dir (bit15) | num)
	  // data = isba rule num

      } else {
        fprintf(stderr, "*** can't find '%s' in %s\n", rule, ipf_conf_name);
	nrules_i_didnt_find++;
      }

    }
    fclose(fd);
    isba_rulenums_available = 1;

    if (nrules_i_didnt_find) {
      fprintf(stderr, "\n    I couldn't find some (%d) ipfilter kernel rules in the\n",
        nrules_i_didnt_find);
      fprintf(stderr,   "    conf. file %s (it only means that I won't be able\n",
        ipf_conf_name);
      fprintf(stderr,   "    to show the isba rule number associated with these rules)\n\n");
    }

    if (0) {
      int group, rule, in_out, key;
      hnode_t * hn;

      group  = 10;
      rule   = 17;
      in_out = 0;

      key = group * 65536 + in_out * 32768 + rule;
      hn = hash_lookup(h_group_num_to_isba_rule, (void *) key);
      fprintf(stderr, "ISBA RULENUM (@%d:%d, %s) = ", group, rule, (in_out ? "out" : "in"));

      if (hn)
        fprintf(stderr, "%d.\n", (int) hnode_get(hn));
      else
        fprintf(stderr, "PAS TROUVE\n");
      exit(1);
    }

}

void add_service_to_table(
  int    proto,
  int    port,
  char * service_name)
{
  hnode_t * hn;

  if (proto == IPPROTO_TCP) {
    if (port < SERV_NAME_MAX)
      tcp_serv_name[port] = strdup(service_name);
    if ((hn = hash_lookup(h_tcp_service_name_to_port, service_name)))
      hnode_put(hn, (void *) port);		// this port was already in table, override it
    else {
      hn = hnode_create((void *) port);
      hash_insert(h_tcp_service_name_to_port, hn, (void *) strdup(service_name));
    }

  } else if (proto == IPPROTO_UDP) {
    if (port < SERV_NAME_MAX)
      udp_serv_name[port] = strdup(service_name);
    if ((hn = hash_lookup(h_udp_service_name_to_port, service_name)))
      hnode_put(hn, (void *) port);		// this port was already in table, override it
    else {
      hn = hnode_create((void *) port);
      hash_insert(h_udp_service_name_to_port, hn, (void *) strdup(service_name));
    }
  }
}


void init_services_names()		// read tcp and udp services names from /etc/services.
{					// some well-known are added if not defined.
  FILE * fd;
  char   buf[BUFSIZ + 1];
  char   name[SMALLSTR];
  char   proto[SMALLSTR];
  int    port, i;
  char * s;

  for (i = 0; i < SERV_NAME_MAX; i++) {
    tcp_serv_name[i] = (char *) NULL;
    udp_serv_name[i] = (char *) NULL;
  }

  h_tcp_service_name_to_port = new_ascii_hash_table("tcp service name to port");
  h_udp_service_name_to_port = new_ascii_hash_table("udp service name to port");

					// some default values in case they're not in /etc/services
  add_service_to_table(IPPROTO_TCP,   20, "ftp-data");		// TCP
  add_service_to_table(IPPROTO_TCP,   22, "ssh");
  add_service_to_table(IPPROTO_TCP,   80, "www");
  add_service_to_table(IPPROTO_TCP,   81, "www-81");
  add_service_to_table(IPPROTO_TCP,   82, "www-82");
  add_service_to_table(IPPROTO_TCP,  110, "pop3");
  add_service_to_table(IPPROTO_TCP,  119, "nntp");
  add_service_to_table(IPPROTO_TCP,  143, "imap");
  add_service_to_table(IPPROTO_TCP,  873, "rsync");
  add_service_to_table(IPPROTO_TCP, 1812, "interscan");
  add_service_to_table(IPPROTO_TCP, 1979, "sc3x");
  add_service_to_table(IPPROTO_TCP, 1987, "patrol-tcp");
  add_service_to_table(IPPROTO_TCP, 7777, "authsrv");

  add_service_to_table(IPPROTO_TCP, 5510, "securidprop");	// SecurId/TCP
  add_service_to_table(IPPROTO_TCP, 5520, "sdlog");
  add_service_to_table(IPPROTO_TCP, 5530, "sdserv");
  add_service_to_table(IPPROTO_TCP, 5540, "sdreport");
  add_service_to_table(IPPROTO_TCP, 5550, "sdadmind");
  
  add_service_to_table(IPPROTO_UDP, 1987, "patrol-udp");	// UDP

  add_service_to_table(IPPROTO_UDP, 1645, "radius");		// SecurId/UDP
  add_service_to_table(IPPROTO_UDP, 1646, "radacct");
  add_service_to_table(IPPROTO_UDP, 5500, "securid");
  add_service_to_table(IPPROTO_UDP, 5540, "sdxauthd");
  for (i = 1024; i < 1048; i++)
    add_service_to_table(IPPROTO_UDP, i, "mprosrv");

  for (i = 0; i < 10; i++) {
    add_service_to_table(IPPROTO_TCP, 6000 + i, "X11");
    add_service_to_table(IPPROTO_TCP, 6010 + i, "X11-ssh");
  }

  if ((fd = fopen("/etc/services", "r"))) {
    while (fgets(buf, BUFSIZ, fd)) {
      if ((s = index(buf, '#')))
        *s = '\0';
      s = buf;
      if (*s) {
        if (3 == sscanf(s, "%s %d/%s ", name, &port, proto)) {
	  if (!strcmp(proto, "tcp"))
	    add_service_to_table(IPPROTO_TCP, port, name);
	  else if (!strcmp(proto, "udp"))
	    add_service_to_table(IPPROTO_UDP, port, name);
        }
      }
    }
  }
}


void init_ipf_module()
{
  int i, j;

  for (i = 0; i < 19; i++)
    for (j = 0; j < 16; j++)
      icmp_type_code[i][j] = (char *) NULL;

  icmp_type_code[ 0][ 0] = "echo-rep";
  icmp_type_code[ 3][ 0] = "net-unr";
  icmp_type_code[ 3][ 1] = "host-unr";
  icmp_type_code[ 3][ 2] = "proto-unr";
  icmp_type_code[ 3][ 3] = "port-unr";
  icmp_type_code[ 3][ 4] = "need-frag";
  icmp_type_code[ 3][ 5] = "src-failed";
  icmp_type_code[ 3][ 6] = "net-unkn";
  icmp_type_code[ 3][ 7] = "host-unkn";
  icmp_type_code[ 3][ 8] = "src-isolat";
  icmp_type_code[ 3][ 9] = "net-prohib";
  icmp_type_code[ 3][10] = "host-prohib";
  icmp_type_code[ 3][11] = "bad-net-tos";
  icmp_type_code[ 3][12] = "bad-host-tos";
  icmp_type_code[ 3][13] = "filter-prohib";
  icmp_type_code[ 3][14] = "host-preced";
  icmp_type_code[ 3][15] = "preced-cutoff";

  icmp_type_code[ 4][ 0] = "src-quench";
  icmp_type_code[ 5][ 0] = "rdr-net";
  icmp_type_code[ 5][ 1] = "rdr-host";
  icmp_type_code[ 5][ 2] = "rdr-tos-net";
  icmp_type_code[ 5][ 3] = "rdr-tos-host";
  icmp_type_code[ 8][ 0] = "echo";
  icmp_type_code[11][ 0] = "timex";
  icmp_type_code[11][ 1] = "timex-reass";
  icmp_type_code[12][ 0] = "bad-ip-header";
  icmp_type_code[12][ 1] = "req-opt-absent";
  icmp_type_code[12][ 2] = "bad-ip-length";

  icmp_type_code[13][ 0] = "tstamp-req";
  icmp_type_code[14][ 0] = "tstamp-rep";
  icmp_type_code[15][ 0] = "info-req";
  icmp_type_code[16][ 0] = "info-rep";
  icmp_type_code[17][ 0] = "adr-mask-req";
  icmp_type_code[18][ 0] = "adr-mask-rep";

  for (i = 0; i < PROTO_NAME_MAX; i++)
    ipproto_name[i] = (char *) NULL;

  ipproto_name[ 0] = "ip";
  ipproto_name[ 1] = "icmp";
  ipproto_name[ 2] = "rip";
  ipproto_name[ 3] = "ggp";
  ipproto_name[ 6] = "tcp";
  ipproto_name[ 8] = "egp";
  ipproto_name[12] = "pup";
  ipproto_name[17] = "udp";
  ipproto_name[20] = "hmp";
  ipproto_name[22] = "xns-idp";
  ipproto_name[27] = "rdp";
  ipproto_name[41] = "ipv6";
  ipproto_name[43] = "ipv6-route";
  ipproto_name[44] = "ipv6-frag";
  ipproto_name[50] = "esp";
  ipproto_name[51] = "ah";
  ipproto_name[58] = "ipv6-icmp";
  ipproto_name[59] = "ipv6-nonxt";
  ipproto_name[60] = "ipv6-opts";

  init_services_names();
  if (ipf_conf_to_hash_table())
    ipfstat_ion_to_hash_table();

}


// ------------------------------------------------------------------------------
// cut the ipfilter log line into fields -> struct logline_fields

void cut_logline_to_ipfilter_fields(logline * ll)
{
  logline_fields * llf;
  char   buf[PADWIDTH];
  char * s;
  char * t;
  char * r;
  int    unknown_format = FALSE;
  int    i, j, k;
  hnode_t * hn;

  llf = (logline_fields *) tmalloc(sizeof(logline_fields));	// create fields structure
  ll->fields = llf;						// attach it to this logline *
  s = ll->line;

  while (! unknown_format) {
    unknown_format = TRUE;			// make "continue" statements exit from loop

    s = (char *) ((int) s + 16);		// skip month, day and syslog time
    skip_to_nth_c (s,            2, ' ');	// skip hostname and 'ipmon[nnnn]:' substring

    while (*(s+2) != ':'			// go straight to the 'HH:MM:SS.uuuuuu' zone
        || *(s+5) != ':'
	|| *(s+8) != '.')
      if (! *s++)				// reached end of line (?!?) => unknown formet, exit.
        continue;

    // fprintf(stderr, "STRING S=%s.\n", s);

    sscanf(s, "%02d:%02d:%02d.%d", &i, &j, &k, &llf->microsec);
    ll->hour   = (u_char) i;			// override syslog time (ipmon time is more accurate)
    ll->minute = (u_char) j;
    ll->second = (u_char) k;
    skip_to_nth_c (s,            1, ' ');	// skip the 'hh:mm:ss.uuuuuu' substring

    if (strchr(s, 'x')				// is there a multiplier ?
     && (int) strchr(s, 'x')< (int) strchr(s, ' ')) {
      copy_to_nth_c(s, buf,      1, ' ');	// yes, get it.
      llf->mult = atoi(buf);
    } else
      llf->mult = 1;

    while (*s == ' ') s++;			// here under OpenBSD there are lots of blanks ...?

    ncopy_to_nth_c(s, llf->intf, 1, ' ', LINTF); // get interface OR "STATE:*" OR "NAT:*")

    if (! strncmp(llf->intf, "STATE:", 6)) {					// -- STATE LINE --
      continue;


    } else if (! strncmp(llf->intf, "NAT:",   4)) {				// -- NAT LINE --
      continue;


    } else {									// -- IPF LINE --

      s++;					// skip '@'
      copy_to_nth_c(s, buf, 1, ':');
      llf->groupnum = atoi(buf);		// get group number
      copy_to_nth_c(s, buf, 1, ' ');
      llf->rulenum = atoi(buf);			// get rule number
      llf->passblock = (*s++ == 'p' ? IPF_PASS : IPF_BLOCK);	// get 'pass | block'
      s++;					

      copy_to_nth_c(s, buf, 1, '.');
      llf->sa1 = atoi(buf);			// get src addr, 1st byte	SOURCE ADDR
      copy_to_nth_c(s, buf, 1, '.');
      llf->sa2 = atoi(buf);			// get src addr, 2nd byte
      copy_to_nth_c(s, buf, 1, '.');
      llf->sa3 = atoi(buf);			// get src addr, 3rd byte

      if (strchr(s, ',')			// will a port be specified ?
       && (int) strchr(s, ',')< (int) strchr(s, ' ')) {
        copy_to_nth_c(s, buf, 1, ',');
        llf->sa4 = atoi(buf);			// get src addr, 4th byte
        copy_to_nth_c(s, buf, 1, ' ');
        llf->sp = atoi(buf);			// get src port

      } else {
        copy_to_nth_c(s, buf, 1, ' ');
        llf->sa4 = atoi(buf);			// get src addr, 4th byte
      }

      if (*s++ != '-') continue;		// skip "->", checking for its presence. If not
      if (*s++ != '>') continue;		// present, "continue" means get out of the false
      s++;					// loop; thus, unknown_format stays to TRUE,
						// and this line will be displayed `raw'
      copy_to_nth_c(s, buf, 1, '.');
      llf->da1 = atoi(buf);			// get dst addr, 1st byte	DEST ADDR
      copy_to_nth_c(s, buf, 1, '.');
      llf->da2 = atoi(buf);			// get dst addr, 2nd byte
      copy_to_nth_c(s, buf, 1, '.');
      llf->da3 = atoi(buf);			// get dst addr, 3rd byte

      if (strchr(s, ',')			// will a port be specified ?
       && (int) strchr(s, ',')< (int) strchr(s, ' ')) {
        copy_to_nth_c(s, buf, 1, ',');
        llf->da4 = atoi(buf);			// get dst addr, 4th byte
        copy_to_nth_c(s, buf, 1, ' ');
        llf->dp = atoi(buf);			// get dst port

      } else {
        copy_to_nth_c(s, buf, 1, ' ');
        llf->da4 = atoi(buf);			// get dst addr, 4th byte
      }

      if (*s++ != 'P') continue;
      if (*s++ != 'R') continue;		// skip "PR"
      s++;

      copy_to_nth_c(s, buf, 1, ' ');
      if      (! strcmp(buf, "tcp"))  llf->protocol = IPPROTO_TCP;
      else if (! strcmp(buf, "udp"))  llf->protocol = IPPROTO_UDP;
      else if (! strcmp(buf, "icmp")) llf->protocol = IPPROTO_ICMP;
      else {
        sscanf(buf, "%d", &i);
	if (i)					// if protocol is specified as its number,
	  llf->protocol = (u_char) i;		// ok, store it
	else					// otherwise this line will be flagged
	  continue;				// as 'unknown format'
      }

      copy_to_nth_c(s, buf, 1, ' ');
      if (strcmp(buf, "len")) continue;		// skip "len"

      copy_to_nth_c(s, buf, 1, ' ');
      llf->headerlen = atoi(buf);		// get ip header length
      copy_to_nth_c(s, buf, 1, ' ');
      llf->packetlen = atoi(buf);		// get packet length

      switch (llf->protocol) {
      case IPPROTO_TCP:
	copy_to_nth_c(s, buf, 1, ' ');		// get TCP flags
	llf->tcp_S      = (index(buf, 'S')  ? 1 : 0);
	llf->tcp_A      = (index(buf, 'A')  ? 1 : 0);
	llf->tcp_P      = (index(buf, 'P')  ? 1 : 0);
	llf->tcp_U      = (index(buf, 'U')  ? 1 : 0);
	llf->tcp_F      = (index(buf, 'F')  ? 1 : 0);
	llf->tcp_R      = (index(buf, 'R')  ? 1 : 0);
	llf->keep_frags = (strstr(s, "K-F") ? 1 : 0);  // get K-F
        break;

      case IPPROTO_ICMP:
	copy_to_nth_c(s, buf, 1, ' ');		// skip "icmp"
	if (strcmp(buf, "icmp")) continue;
	copy_to_nth_c(s, buf, 1, '/');
	llf->icmp_type = atoi(buf);		// get icmp type
	copy_to_nth_c(s, buf, 1, ' ');
	llf->icmp_code = atoi(buf);		// get icmp code

	if (! strncmp(s, "for", 3)) {		// is a "for ..." (orig pkt) specified ?
	  if (! (r = strstr(s, "IN")))
	    if (! (r = strstr(s, "OUT")))	// make r point to the end of the orig pkt spec.
	      continue;
	  for (t = buf; *s && ((int) s < (int) r); )
	    *t++ = *s++;			// copy orig pkt spec to buf
	  *t = '\0';
	  llf->icmp_for = tstrdup(buf);		// copy buf to a mallocated string for llf
	} else {				// s points to "IN" or "OUT".
	  llf->icmp_for = (char *) NULL;
	}
        break;
      }

      llf->keep_state = (strstr(s, "K-S") ? 1 : 0);	// get K-S
      if      (strstr(s, "IN"))  llf->in_out = 0;	// get IN | OUT
      else if (strstr(s, "OUT")) llf->in_out = 1;
      else continue;

      if (llf->groupnum != -1				// ISBA RULE NUM
       && isba_rulenums_available
       && ll->emitted_by_local_host) {
	i = (int) (llf->groupnum << 16) + (llf->in_out ? 32768 : 0) + llf->rulenum;
	if ((hn = hash_lookup(h_group_num_to_isba_rule, (void *) i))) {
	  llf->isbarule = (int) hnode_get(hn);

	} else
	  llf->isbarule = -1;

      } else
	llf->isbarule = -1;

    }

    unknown_format = FALSE;			// if I'm here, then parsing succeeded.
    break;
  }

  llf->unknown_format = unknown_format;

}

char hide_char(
  char c)
{
  if (c >= 'a' && c <= 'z') {
    c += (char) (option_d_disguise_addr && 0xf);
    if (c > 'z')
      c -= ('z' - 'a');
  }

  return c;
}

void hide_ip_address(
  u_char * pa1,
  u_char * pa2,
  u_char * pa3,
  u_char * pa4)
{
  u_char rnd1, rnd2, rnd3, rnd4;

  rnd1 = (u_char)               (option_d_disguise_addr        & 0xff);
  rnd2 = (u_char) rnd1 ^ ((int) (option_d_disguise_addr >> 8)  & 0xff);
  rnd3 = (u_char) rnd2 ^ ((int) (option_d_disguise_addr >> 16) & 0xff);
  rnd4 = (u_char) rnd3 ^ ((int)  option_d_disguise_addr >> 24);

  if (*pa1 >= 224		// multicast or broadcast addr: leave as is
   || (*pa1 == 0))		// 1st byte null (?!): leave as is
    return;

  // if the address is a rfc1918 one, or broadcast/localhost/0 :
  // don't change the first two bytes.

  if (! ((*pa1 == 10)					// 10/8
      || (*pa1 == 192 && *pa2 == 168)			// 192.168/16
      || (*pa1 == 172 && (*pa2 & 0xf0) == 16))		// 172.16/24
      || (*pa1 == 255)					// 255/8
      || (*pa1 == 127)					// 127/8
      || (*pa1 == 0)) {					// 0/8

    *pa1 = *pa1 ^ rnd1;
    if (*pa1 == 0 || *pa1 == 255)
      *pa1 = rnd2;

    *pa2 = *pa2 ^ rnd1;
    if (*pa2 == 0 || *pa2 == 255)
      *pa2 = rnd3;
  }

  // if broadcast/localhost/0 : don't change the last two bytes.

  if (! (*pa1 == 255)					// 255/8
      || (*pa1 == 127)					// 127/8
      || (*pa1 == 0)) {					// 0/8

    *pa3 = *pa3 ^ rnd3;
    if (*pa3 == 0 || *pa3 == 255)
      *pa3 = rnd4;
  
    *pa4 = *pa4 ^ rnd4;
    if (*pa4 == 0 || *pa4 == 255)
      *pa4 = rnd1;
  }
}


// ------------------------------------------------------------------------------
// prepare the chtype * string from the computed fields

void ipmon_fields_to_fmt_and_attr_lines(
  logline * ll)
{
  logline_fields * llf = ll->fields;
  char     buf[PADWIDTH + 1];
  char     buf2[SMALLSTR];
  char     buf3[SMALLSTR];
  char     fmtline[PADWIDTH + 1];
  char     attrline[PADWIDTH + 1];
  char *   cur_fmt_char;
  char *   cur_attr_char;
  int      is_tcp, is_udp, is_icmp;
  int      mark_attr, attr, port;
  u_char   sa1, sa2, sa3, sa4, da1, da2, da3, da4;

    mark_attr = (ll->is_marked ? MYA_REVERSE : MYA_NORMAL) | MY_BIT7;

    is_tcp  = (llf->protocol == IPPROTO_TCP);
    is_udp  = (llf->protocol == IPPROTO_UDP);
    is_icmp = (llf->protocol == IPPROTO_ICMP);

    *fmtline  = '\0';
    *attrline = '\0';
    cur_fmt_char  = fmtline;
    cur_attr_char = attrline;

    if (option_show_date) {						// DATE
      sprintf(buf, "%3s %2d ", ll->month, ll->day);
      append_to_fmt_and_attr_strings(cur_fmt_char, cur_attr_char, buf, mark_attr);
    }

    if (option_show_time) {						// TIME
      sprintf(buf, "%02d:%02d:%02d", ll->hour, ll->minute, ll->second);
      append_to_fmt_and_attr_strings(cur_fmt_char, cur_attr_char, buf, mark_attr);
    }

    if (option_show_microsec) {						// MICROSEC
      sprintf(buf, ".%06d", llf->microsec);
      append_to_fmt_and_attr_strings(cur_fmt_char, cur_attr_char, buf, mark_attr);
    }

    if ((option_show_time || option_show_microsec) && option_show_hostname) {
      *cur_fmt_char++  = ' ';						// if time or microseconds,
      *cur_attr_char++ = mark_attr;					// and hostname, are displayed,
    }									// insert a blank

    if (option_show_hostname) {						// HOSTNAME
      sprintf(buf, "%-*.*s ", nchars_host, nchars_host, ll->host);
      if (option_d_disguise_addr) {					// if hide mode, replace the first 2 letters
        buf[0] = hide_char(buf[0]);
        buf[1] = hide_char(buf[1]);
        buf[2] = hide_char(buf[2]);
      }
      append_to_fmt_and_attr_strings(cur_fmt_char, cur_attr_char, buf, mark_attr);
    }

    if (llf->mult > 1)
      sprintf(buf, "%3dx ", llf->mult);					// MULTIPLIER
    else
      sprintf(buf, "     ");
    append_to_fmt_and_attr_strings(cur_fmt_char, cur_attr_char, buf, mark_attr);

    if (option_show_intf) {						// INTERFACE
      sprintf(buf, "%4s ", llf->intf);
      append_to_fmt_and_attr_strings(cur_fmt_char, cur_attr_char, buf, mark_attr);
    }

    if (option_show_rulenum) {						// RULE GROUP:NUM
      if (llf->groupnum == -1)
           sprintf(buf, "         ");
      else sprintf(buf, "%5d:%-3d", llf->groupnum, llf->rulenum);
      append_to_fmt_and_attr_strings(cur_fmt_char, cur_attr_char, buf, mark_attr);
    }

    if (option_show_isbanum) {						// ISBA RULE NUM
      if (llf->groupnum == -1 || llf->isbarule < 1 || ! isba_rulenums_available)
           sprintf(buf, "     ");
      else sprintf(buf, " %3d ", llf->isbarule);
      append_to_fmt_and_attr_strings(cur_fmt_char, cur_attr_char, buf, mark_attr);
    }
 
    sprintf(buf, "%c ", (llf->passblock == IPF_PASS ? 'p' : 'b'));	// PASS/BLOCK
    append_to_fmt_and_attr_strings(cur_fmt_char, cur_attr_char, buf,
      mark_attr | (llf->passblock == IPF_PASS ? MYA_NORMAL : MYA_BOLD));

    sa1 = llf->sa1;
    sa2 = llf->sa2;
    sa3 = llf->sa3;
    sa4 = llf->sa4;
    da1 = llf->da1;
    da2 = llf->da2;
    da3 = llf->da3;
    da4 = llf->da4;

    if (option_d_disguise_addr) {
      hide_ip_address(&sa1, &sa2, &sa3, &sa4);
      hide_ip_address(&da1, &da2, &da3, &da4);
    }

    if (option_sparse_ip_addr) {					// ALIGN IP ADDR BYTES
      if (is_tcp || is_udp)
        sprintf(buf, "%3d.%3d.%3d.%3d %-5d -> %3d.%3d.%3d.%3d %-5d",
	  sa1, sa2, sa3, sa4, llf->sp,
	  da1, da2, da3, da4, llf->dp);
      else
        sprintf(buf, "%3d.%3d.%3d.%3d       -> %3d.%3d.%3d.%3d      ",
	  sa1, sa2, sa3, sa4,
	  da1, da2, da3, da4);

    } else {								// COMPACT IP ADDR BYTES
      sprintf(buf2, "%d.%d.%d.%d", sa1, sa2, sa3, sa4);
      if (is_tcp || is_udp)
           sprintf(buf, "%15s %-5d -> ", buf2, llf->sp);
      else sprintf(buf, "%15s       -> ", buf2);

      sprintf(buf3, "%d.%d.%d.%d", da1, da2, da3, da4);
      if (is_tcp || is_udp)
           sprintf(buf2, "%15s %-5d", buf3, llf->dp);
      else sprintf(buf2, "%15s      ", buf3);

      strcat(buf, buf2);
    }
    append_to_fmt_and_attr_strings(cur_fmt_char, cur_attr_char, buf, mark_attr);

    if (llf->protocol < PROTO_NAME_MAX && ipproto_name[llf->protocol])
      sprintf(buf, " %-4s ", ipproto_name[llf->protocol]);
    else
      sprintf(buf, " #%-3d ", llf->protocol);
    append_to_fmt_and_attr_strings(cur_fmt_char, cur_attr_char, buf, mark_attr);

    if (option_show_headerlen) {
      sprintf(buf, "%2d ", llf->headerlen);
      append_to_fmt_and_attr_strings(cur_fmt_char, cur_attr_char, buf, mark_attr);
    }
    sprintf(buf, "%4d ", llf->packetlen);
    append_to_fmt_and_attr_strings(cur_fmt_char, cur_attr_char, buf, mark_attr);

    attr = mark_attr | (llf->passblock == IPF_PASS ? MYA_NORMAL : MYA_BOLD);

    if (llf->protocol == IPPROTO_TCP) {
      *cur_fmt_char++  = (llf->tcp_S ? 'S' : '.');
      *cur_attr_char++ = (llf->tcp_S ? MYA_BOLD : MYA_NORMAL) | mark_attr;
      *cur_fmt_char++  = (llf->tcp_A ? 'A' : '.');
      *cur_attr_char++ = mark_attr;
      *cur_fmt_char++  = (llf->tcp_P ? 'P' : '.');
      *cur_attr_char++ = mark_attr;
      *cur_fmt_char++  = (llf->tcp_U ? 'U' : '.');
      *cur_attr_char++ = mark_attr;
      *cur_fmt_char++  = (llf->tcp_F ? 'F' : '.');
      *cur_attr_char++ = (llf->tcp_F ? MYA_BOLD : MYA_NORMAL) | mark_attr;
      *cur_fmt_char++  = (llf->tcp_R ? 'R' : '.');
      *cur_attr_char++ = (llf->tcp_R ? MYA_BOLD : MYA_NORMAL) | mark_attr;

      sprintf(buf, " %2s %2s %3s ",
        (llf->keep_frags ? "KF"  : ""),
	(llf->keep_state ? "KS"  : ""),
	(llf->in_out     ? "OUT" : "IN"));
      append_to_fmt_and_attr_strings(cur_fmt_char, cur_attr_char, buf, mark_attr);

      // try to guess the service involved:
      // - 1. if it's a SYN packet, the service is the destination port
      // - 2. if I know only one of the ports' service names, I take this one as the service
      // - 3. otherwise I take the smallest of the src and the dst ports.

      if (llf->tcp_S && ! (llf->tcp_A || llf->tcp_F || llf->tcp_R))	// 1.
        port = llf->dp;
      else if (llf->sp < SERV_NAME_MAX && tcp_serv_name[llf->sp]
         && ! (llf->dp < SERV_NAME_MAX && tcp_serv_name[llf->dp]))	// 2.a
	port = llf->sp;
      else if (llf->dp < SERV_NAME_MAX && tcp_serv_name[llf->dp]
         && ! (llf->sp < SERV_NAME_MAX && tcp_serv_name[llf->sp]))	// 2.b
	port = llf->dp;
      else								// 3.
        port = min(llf->sp, llf->dp);

      if (port < SERV_NAME_MAX && tcp_serv_name[port])
           sprintf(buf, "(%s)", tcp_serv_name[port]);
      else sprintf(buf, "(%d)", port);

      append_to_fmt_and_attr_strings(cur_fmt_char, cur_attr_char, buf, attr);

    } else if (llf->protocol == IPPROTO_UDP) {

      sprintf(buf, "          %2s %3s ",
        (llf->keep_state ? "KS" : ""),
	(llf->in_out     ? "OUT" : "IN"));
      append_to_fmt_and_attr_strings(cur_fmt_char, cur_attr_char, buf, mark_attr);

      // try to guess the service involved:
      // - 1. if I know only one of the ports' service names, I take this one as the service
      // - 2. otherwise I take the smallest of the src and the dst ports.

      if   (llf->sp < SERV_NAME_MAX && udp_serv_name[llf->sp]
      && ! (llf->dp < SERV_NAME_MAX && udp_serv_name[llf->dp]))		// 1.a
	port = llf->sp;
      else if (llf->dp < SERV_NAME_MAX && udp_serv_name[llf->dp]
         && ! (llf->sp < SERV_NAME_MAX && udp_serv_name[llf->sp]))	// 1.b
	port = llf->dp;
      else								// 2.
        port = min(llf->sp, llf->dp);

      if (port < SERV_NAME_MAX && udp_serv_name[port])
           sprintf(buf, "(%s)", udp_serv_name[port]);
      else sprintf(buf, "(%d)", port);

      append_to_fmt_and_attr_strings(cur_fmt_char, cur_attr_char, buf, attr);

    } else if (llf->protocol == IPPROTO_ICMP) {
      sprintf(buf, "  %2d/%1d    %2s %3s ",
        llf->icmp_type,
	llf->icmp_code,
	(llf->keep_state ? "KS" : ""),
	(llf->in_out     ? "OUT" : "IN"));
      append_to_fmt_and_attr_strings(cur_fmt_char, cur_attr_char, buf, mark_attr);

      if (llf->icmp_type > 18
       || llf->icmp_code > 15
       || ! icmp_type_code[llf->icmp_type][llf->icmp_code]) {		// (TYPE/CODE NAME: ICMP)
        sprintf(buf, "(invalid type/code ?!?)");
        append_to_fmt_and_attr_strings(cur_fmt_char, cur_attr_char, buf, mark_attr ^ MYA_REVERSE);
      } else {
        sprintf(buf, "(%s)", icmp_type_code[llf->icmp_type][llf->icmp_code]);
        append_to_fmt_and_attr_strings(cur_fmt_char, cur_attr_char, buf, attr);
      }

      if (llf->icmp_for) {						// ICMP "FOR" PACKET
        sprintf(buf, "  %s", llf->icmp_for);
        append_to_fmt_and_attr_strings(cur_fmt_char, cur_attr_char, buf, mark_attr);
      }
    }

/*
    for (cs = chstr, i = 0; *cs; cs++)		// make cs point to the trailing '\0'
      i++;					// (set i = length(chstr))
    for ( ; i < PADWIDTH - 1; i++)		// pad with blanks till PADWIDTH chars
      *cs++ = (chtype) ' ' | mark_attr;
    *cs = 0;
*/

    *cur_fmt_char  = '\0';			// terminate formatted (both chars and attr) strings
    *cur_attr_char = '\0';

    if (ll->fmtline && ll->fmtline != ll->line)
      tfree(ll->fmtline);
    if (ll->attrline)
      tfree(ll->attrline);

    ll->fmtline  = tstrdup(fmtline);
    ll->attrline = tstrdup(attrline);

    // if "compact memory" enabled, free the ll->line string
    // if (option_z_compress)
    //   tfree(ll->line);
}


// ------------------------------------------------------------------------------

chtype * ipf_headers_line(chtype * chstr)
{
  int    i;

  *chstr = (chtype) 0;

  if (option_show_date)
    strcat_to_chstr(chstr, A_REVERSE, "%-6s ", "Date");

  if (option_show_time)
    strcat_to_chstr(chstr, A_REVERSE, "%-*s ", (option_show_microsec ? 14 : 7), "Time");
  else if (option_show_microsec)
    strcat_to_chstr(chstr, A_REVERSE, "%-7s", "Time");

  if (option_show_hostname)
    strcat_to_chstr(chstr, A_REVERSE, "%s%-*s ",
      (option_show_time || option_show_microsec ? " " : ""),
      nchars_host, "Hostname");

  strcat_to_chstr(chstr,   A_REVERSE, " %-3s ", "Mul");
  if (option_show_intf)
    strcat_to_chstr(chstr, A_REVERSE, "%-4s ", "Intf");
  if (option_show_rulenum)
    strcat_to_chstr(chstr, A_REVERSE, "%-7s ", "Rulenum ");
  if (option_show_isbanum)
    strcat_to_chstr(chstr, A_REVERSE, "%-4s ", "Isba");

  strcat_to_chstr(chstr,   A_REVERSE, "pb %-20s -> %-20s %-5s",
    "Source-Address Port", "   Dest-Address Port", "Proto ");

  if (option_show_headerlen)
    strcat_to_chstr(chstr, A_REVERSE, "HL ");

  strcat_to_chstr(chstr, A_REVERSE, "PktL  Flags/Type  I/O Service");

  i = screen_maxx - chstrlen(chstr) - strlen(logfile) - strlen(local_host_name) - 2;
  if (i > 4) {								// try to display hostname:/path/to/logfile
    strcat_to_chstr(chstr, A_REVERSE, "%*s%s:%s",			// in the headers line
		    i, "", local_host_name, logfile);
  } else {
    i = screen_maxx - chstrlen(chstr) - strlen(local_host_name) - 1;	// if there's not enough room, try to display
    if (i > 2)								// only the hostname
      strcat_to_chstr(chstr, A_REVERSE, "%*s%s", i, "", local_host_name);
  }

  strcat_to_chstr(chstr, A_REVERSE, "%*s", PADWIDTH - chstrlen(chstr) - 1, "");

  return chstr;
}


// ------------------------------------------------------------------------------


void tests()
{
  logline * ll;
  logline_fields * llf;

  ll = (logline *) tmalloc(sizeof(logline));
  ll->line = strdup(
  "Jun 27 22:35:13 cube ipmon[32240]: 22:35:12.313846             tun0 @10:19 b 29.6.1.49 -> 6.4.2.20 PR icmp len 20 84 icmp 8/0 IN");
  ll->line = strdup(
  "Jun 29 08:19:37 proxfed1 ipmon[26364]: 08:19:37.034022 hme0 @0:17 p 192.168.1.208,37240 -> 192.168.1.81,53 PR udp len 20 71  K-S OUT");

  msg("line = %s", ll->line);
  cut_logline_to_ipfilter_fields(ll);
  llf = (logline_fields *) ll->fields;
  fprintf(stderr, "ll->fields->unknown_format = %d.\n", llf->unknown_format);

  exit(0);
}

