/*
 * readlog.c
 *
 * Curses based IP-Filter logs viewer
 *
 * Copyright (C) 2001  Pierre Berthomier
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2
 * as published by the Free Software Foundation, provided that this notice
 * is preserved in its entirety and due credit is given to the original
 * author and the contributors.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#include "vlog.h"

/* start_logfiles_reading_threads()
 *
 * o read the last LASTNBYTES bytes from logfile to be able
 *   to display the last log lines immediately
 * o create a thread that will read log lines as they arrive
 * o create a thread that will read log lines from
 *   (around) LASTNBYTES before the end to the beginning of the
 *   current logfile, and that will read (backwards also)
 *   the previous rotated logfiles.
 *
 * - pointer to first `struct logline' (head_line) is initialized,
 * - pointer to last `struct logline' (tail_line) too, and this one goes on running.
 */

char oldest_date_seen[SMALLSTR];

void start_logfiles_reading_threads(char * logfile)
{
  FILE *	fd_tail;
  long		file_size;
  int		nbytes2read, nbytes_read;
  char *	buffer = NULL;	// to avoid gcc warning
  struct stat	statbuf;
  int           ret;

  *oldest_date_seen = '\0';

  /* Read last log lines and start the 'tail -f' thread */

  if ((fd_tail = fopen(logfile, "r"))) {

    stat(logfile, &statbuf);
    file_size = (long) statbuf.st_size;
    nbytes2read = min(file_size, LASTNBYTES);
    fseek(fd_tail, -nbytes2read, SEEK_END);		// seek(end - nbytes2read)
    							// (read the end now to show it immediately)
    buffer = (char *) tmalloc(sizeof(char) * (LASTNBYTES + 1));
    nbytes_read = fread(buffer, 1, nbytes2read, fd_tail); // read `nbytes2read' bytes

    if (nbytes_read == nbytes2read) {
      // msg("READ %d bytes", nbytes_read);
      buffer[nbytes_read] = '\0';

      if (file_size - nbytes2read)			// unless fseek pointed to start of file,
        while (*(buffer++) != '\n')			// I'm probably in the middle of a line.
          nbytes_read--;				// look for the next newline char

      append_buffer_to_loglines_queue(buffer);		// put those lines in my loglines structure
      reading_1st_loglines = 0;
      if (! option_start_month				// if no start date was given on command line,
       && option_tail_nnn_f == -1			// and if option -<nn>f or -n <nn> were not given,
       && option_n_lines_only == -1)
        cur_bottom_ll = new_tail_line;			// initialize the currently displayed bottom log line ptr
							// for main thread to display lines immediately
      initialize_line_numerotation();			// give linenumbers to lines read

      ret = pthread_create(&tail_f_thread_id , NULL,	// give control to a `tail -f' thread
			   (void *) tail_f_thread, (void *) fd_tail);
      if (ret)
        err("couldn't start 'tail -f' thread (return code %d)", ret);

      youngest_logfile_start_offset =			// position from which I've read lines
        (long) file_size - nbytes_read;

    } else
      err("start_logfiles_reading_threads(): read only %d bytes instead of %d ?!?",
          nbytes_read, nbytes2read);
  } else
    err_perr("can't read %s", logfile);


  /* Start the 'read backwards' thread */

    ret = pthread_create(&reverse_reading_thread_id , NULL,	   // start reverse logfile reading
			 (void *) reverse_reading_thread, &youngest_logfile_start_offset);
    if (ret)
      err("couldn't start 'reverse reading' thread (return code %d)", ret);
}

// -------------------------------------------------------------------------------------

/* tail_f_thread()
 *
 * the thread that append `struct logline' entries to the queue as soon as they arrive.
 * Also periodically check if the log file was rotated, and if so, reopen it.
 */

void tail_f_thread(
  FILE * fd_tail)
{
  int   	nbytes_read;
  char *	buffer;
  struct stat	statbuf;
  long		logfile_lastsize = 0;
  int		count = 0;
  int		ret;
  logline *	ll;

  buffer = (char *) tmalloc(TAIL_F_READ_BLOCK + 1);

  while (1) {
    nbytes_read = fread(buffer, 1, TAIL_F_READ_BLOCK, fd_tail);

    if (nbytes_read > 0) {			// some line(s) arrived
      buffer[nbytes_read] = '\0';

      append_buffer_to_loglines_queue(buffer);

    } else {
      if (feof(fd_tail)) {
        clearerr(fd_tail);			// nothing new in log file. sleep.
	napms(TAIL_F_THREAD_SLEEP_MS);

      } else if (errno && errno != EINTR && errno != EAGAIN)	// EINTR is generated on a WINCH
        err_perr("error on reading %s (%d)", logfile, errno);	// other error condition.

      // check if the logfile has rotated
      // every (CHK_ROTATED_LOGFILE_COUNT * TAIL_F_THREAD_SLEEP_MS) millisec

      if (count++ > CHK_ROTATED_LOGFILE_COUNT || ! logfile_lastsize) {
	ret = stat(logfile, &statbuf);
        if (ret && errno != ENOENT)
	  err_perr("tail_f_thread: couldn't stat %s", logfile);

        if ((ret && errno == ENOENT)					// if logfile disappeared
	 || (logfile_lastsize && logfile_lastsize > statbuf.st_size)) {	// or its size decreased
									// => then it was rotated.
          fclose(fd_tail);

          count = 10;							// try to reopen the
          while (count && ! (fd_tail = fopen(logfile, "r"))) {		// new logfile during 10 sec
            count--;
            sleep(1);
          }
          if (! fd_tail)
            err_perr("logfile rotated, but I couldn't reopen it");

          ll = create_logline(logfile,				// insert a separator line with the
            NULL, tail_line);					// name of the logfile
	  if (tail_line)
            tail_line->next = ll;
	  else {
	    head_line = ll;
	    tail_line = ll;
	    initialize_line_numerotation();
	  }
          new_tail_line = ll;
          ll->is_a_separator = 1;
	  new_lines_arrived = 1;
        }

	if (! ret)
          logfile_lastsize = (long) statbuf.st_size;		// update logfile last size
        count = 0;						// and check counter
      }
    }
  }
}

// -----------------------------------------------------------------------------------------------

/* reverse_reading_thread()
 *
 * Read logfile (/var/log/ipflog) from the point near the end where
 * start_logfiles_reading_threads() started, backwards to the start.
 * Then open rotated logfiles (/var/log/ipflog.*), also in reverse order.
 * Create (logline *) structures for each line read.
 */

void reverse_reading_thread(
  long * logfile_start_offset)
{
  FILE *	fd_back;
  int   	nbytes_read;
  int   	nbytes2read;
  char *	buffer;
  char *	sub_buffer;
  int   	reading_logfile_number = -1;	// -1 = /var/log/ipflog, 0 = /var/log/ipflog.0, ...
  char *	reading_logfile_name;
  struct stat	statbuf;
  logline *	ll;
  long		logfile_offset = *logfile_start_offset;

  reading_logfile_name = (char *) tmalloc(SMALLSTR + 1);
  buffer	       = (char *) tmalloc(BACKWARDS_READ_BLOCK + 1);
  sprintf(reading_logfile_name, "%s", logfile);

  strcpy(reverse_thread_status, "Initializing");	// will be displayed in status line

  sprintf(buffer,
    "/bin/ls -1 %s.* 2>/dev/null | awk -F. '{if ($NF>M) M=$NF} END {printf(\"%%d\\n\", M)}'",
    logfile);

  if ((old_logfiles_fd = popen(buffer, "r"))) {	// read the number of oldest logfile (ipflog.12 --> 12)
    napms(100);
    if (fgets(buffer, 256, old_logfiles_fd)) {
      pclose(old_logfiles_fd);
      old_logfiles_fd = NULL;

      if (*buffer == '\n')
        number_of_oldest_logfile = -1;
      else
        sscanf(buffer, "%d", &number_of_oldest_logfile);

    } else
      err_perr("couldn't read the number of old logfiles ?!");
  } else
    err_perr("couldn't get the number of old logfiles ?!");

  while (reading_logfile_number <= number_of_oldest_logfile && ! reverse_thread_must_stop) {

    if (! logfile_offset) {				// if we read all lines from current logfile, open next

      if (mode_curses) {				// if we're in curses mode,
        ll = create_logline(reading_logfile_name,	// first insert a separator line
          head_line, NULL);				// with the name of the logfile
        if (head_line)					// if queue exists, prepend this line
          head_line->prev = ll;				// at its head
        else {						// otherwise initialize queue tail pointer and
          tail_line = ll;					// and (just below) queue head pointer
          initialize_line_numerotation();
        }
        head_line = ll;
        ll->is_a_separator = 1;
      }

      if (! option_r_rotated)				// if option -r (rotated) is not set,
        break;						//   I don't read previous logfiles.
      if (++reading_logfile_number > number_of_oldest_logfile)
        break;						// it was the oldest logfile ? exit from while loop.

      sprintf(reading_logfile_name, "%s.%d", logfile,
        reading_logfile_number);
      stat(reading_logfile_name, &statbuf);		// reinitialize start offset = end of file
      logfile_offset = (long) statbuf.st_size;
    }

    if ((fd_back = fopen(reading_logfile_name, "r"))) {

      while (logfile_offset > 0 && ! reverse_thread_must_stop) {
	napms(BACKWARDS_THREAD_SLEEP_MS);

        nbytes2read = min(logfile_offset, BACKWARDS_READ_BLOCK);// read at most BACKWARDS_READ_BLOCK
        fseek(fd_back, logfile_offset - nbytes2read, SEEK_SET); // seek(offset - nbytes2read)

        nbytes_read = fread(buffer, 1, nbytes2read, fd_back);
        if (nbytes_read > 0) {
	  sub_buffer = buffer;
	  if (logfile_offset - nbytes2read)		// unless fseek pointed to start of file,
            while (*(sub_buffer++) != '\n')		// I'm probably in the middle of a line.
              nbytes_read--;				// look for the next newline char

          sub_buffer[nbytes_read] = '\0';

	  // msg("READ %d:%d", (int) logfile_offset - nbytes2read, nbytes_read);

          prepend_buffer_to_loglines_queue(sub_buffer);
	  logfile_offset = logfile_offset - nbytes_read;

        } else if (errno)
          err_perr("reverse_reading_thread(): error on reading %s:%d-%d (errno %d)",
	    reading_logfile_name, logfile_offset - nbytes2read, logfile_offset, errno);
      }
    }
  }

  if (! cur_bottom_ll && ! i_found_start_logline) {	// if a start date/time was given on command line,
    go_to_top_of_logs();				// and if I didn't see this date/time yet: go to top
    i_found_start_logline = 1;
  }

  if (! *oldest_date_seen && head_line) {
    sprintf(oldest_date_seen, "%3s %d", head_line->month, head_line->day);
  }

  sprintf(reverse_thread_status, "Start: %s", oldest_date_seen);	// update status indicator
  new_backw_lines_read = 1;				// and tell main thread to update status line
  reverse_thread_must_stop = 1;
}


// -----------------------------------------------------------------------------------------------


logline * create_logline(
  char *    raw_line,
  logline * next,
  logline * prev)
{
  logline * ll;
  char      month[256];
  char      buffer[BUFSIZ];
  int       day, hour, minute, second, year;
  char *    s;
  char *    t;
  char *    u;

  ll = (logline *) tmalloc(sizeof(logline));		// create new `logline' to hold this line
  if (strlen(raw_line) < PADWIDTH - 1)
    ll->line = tstrdup(raw_line);
  else {
    ll->line = tmalloc(PADWIDTH - 1);			// very long lines ( > PADWIDTH - 1) are truncated
    strncpy(ll->line, raw_line, PADWIDTH - 1);
    *(ll->line + PADWIDTH - 1) = '\0';
  }

  ll->next = next;
  ll->prev = prev;

  ll->is_marked		= 0;
  ll->matched		= 0;
  ll->is_a_separator	= 0;
  ll->was_printed	= 0;
  ll->not_an_ipmon_line	= (strstr(ll->line, "ipmon[") ? 0 : 1);
  ll->emitted_by_local_host = 0;
  ll->is_before_start_date = 0;
  ll->serial		= ((u_char) cur_fmt_serial - 1);
  ll->srch_serial	= ((u_char) cur_srch_serial - 1);
  ll->pid		= 0;
  ll->fields		= (void *) 0;
  ll->fmtline		= (char *) NULL;
  ll->attrline		= (char *) NULL;

  if      (ll->prev && ll->prev->linenum != MAXINT)	// give this line its line number
    ll->linenum	= ll->prev->linenum + 1;		// if previous one's ou next one's
  else if (ll->next && ll->next->linenum != MAXINT)	// number is known
    ll->linenum	= ll->next->linenum - 1;
  else
    ll->linenum	= MAXINT;				// MAXINT == uninitialized line number

  for (s = ll->line; *s; s++)				// replace each Tab with a space
   if (*s == '\t')					// (Tabs scramble curses display)
     *s = ' ';

  ll->day = 0;						// these are for unknows format lines.
  ll->month[0] = '\0';					// set these default values now,
  ll->message = (char *) NULL;				// then I'll look if it is a logline format I know.

							// [A] IS THIS A STANDARD SYSLOG FORMAT ? -------------------------
  if (5 == sscanf(ll->line, "%3s %2d %2d:%2d:%2d", month, &day, &hour, &minute, &second)) { // (beginning with date)
    strncpy(ll->month, month, 4);
    ll->day	= (u_char) day;
    ll->hour	= (u_char) hour;
    ll->minute	= (u_char) minute;
    ll->second	= (u_char) second;

    s = (char *) ((int) ll->line + 16);		// skip month, day and time
    ncopy_to_nth_c(s, buffer, 1, ' ', BUFSIZ);	// get name of host that issued the message
    ll->host = tstrdup(buffer);
    if (! strcmp(buffer, local_host_name))
      ll->emitted_by_local_host = 1;

    if (reading_1st_loglines && ! ll->not_an_ipmon_line)
      mode_ipmon = 1;

    if (ll->not_an_ipmon_line && ! mode_ipmon) {

      if ((t = strchr(s, ':'))			// IF message contains a ':',
       && ((int) t - (int) s) < 40) {		// AND if it is within the 1st 40 chars of message,

	if ((u = strchr(s, '['))		// AND if message contains a '[' before next ':',
	 && (int) u < (int) t) {		// THEN it means that a [pid] is specified.
          ncopy_to_nth_c(s, buffer, 1, '[', BUFSIZ);	// get name of process that issued the message
	} else {
          ncopy_to_nth_c(s, buffer, 1, ':', BUFSIZ);	// get name of process that issued the message
	}
	
        ll->fields = (logline_fields *) tstrdup(buffer);// use the `fields' pointer to remember proc. name
	if (reading_1st_loglines && strlen(buffer) > proc_name_width)	// if reading 1st lines of logfile,
	  proc_name_width = strlen(buffer);					// compute the max of "process names" widths

	if (u && (int) u < (int) t) {		// if a [pid] is specified, get it now.
	  sscanf(s, "%d", &ll->pid);
	  skip_to_nth_c(s, 1, ':');		// advance s till next ':' (after the "[pid]:")
	}
	if (*s == ' ')
	  s++;					// also skip first blank if any
      }

      ll->message = s;				// ll->message points after hostname
    }						// (used for 'plain syslog' mode)

    if (option_start_month			// if a start date was given on command line, compare it
     && compare_ll_date(ll, option_start_month, option_start_day,
			    option_start_hour, option_start_minute, option_start_second) == -1)
      ll->is_before_start_date = 1;

/*
    if (ll->is_before_start_date)
      fprintf(stderr, "BEFORE: %s\n", ll->line);
    else
      fprintf(stderr, " AFTER: %s\n", ll->line);
*/

  } else {

    s = (char *) ll->line;
    u = s;
    skip_to_nth_c(u, 3, ' ');

    // debug_timestamp();
    // printd1("s = %s\n", s);

    if (*u == '[' && *(u+3) == '/' && *(u+7) == '/') {	// [B] IS THIS AN HTTPD ACCESS LOG LINE ? ---------------------------
      copy_to_nth_c(s, buffer, 1, ' ');			// (4th field = [01/Nov/2001:23:50:05 -0500])
      ll->client_addr = tstrdup(buffer);		// get HTTP client IP address or FQDN

      s = s + 2;					// skip the ' - '
      copy_to_nth_c(s, buffer, 1, ' ');			// get the HTTP user if any
      if (buffer[0] == '-') {
        ll->client_user = (char *) NULL;
      } else {
        ll->client_user = tstrdup(buffer);
      }

      printd1("HTTPD ACCESS logline: %s.\n", s);

      if (*s == '['				// ensure I'm at the beginning of the date ("[01/Nov/2001:23:50:05 -0500]")
       && (6 == sscanf(s + 1, "%d/%3s/%d:%d:%d:%d",
                      &day, month, &year, &hour, &minute, &second))) {
	ll->day		= (u_char) day;
	ll->hour	= (u_char) hour;
	ll->minute	= (u_char) minute;
	ll->second	= (u_char) second;
	strncpy(ll->month, month, 4);

	skip_to_nth_c(s, 1, ']');		// skip the date
	s++;					// skip the blank. I'm on the '"' before the URL.
	ll->message = s;			//              => this is the 'message'.
	skip_to_nth_c(s, 2, '"');		// skip the URL (two '"')
	s++;					// skip the blank. I'm on the HTTP RETURN CODE.
	if (2 != sscanf(s, "%d %d", &ll->http_code, &ll->http_size)) {
	  printd1("SSCANF HTTP CODE, SIZE retourne %d.\n", sscanf(s, "%d %d", &ll->http_code, &ll->http_size));
	  ll->http_code = 0;
	  ll->http_size = 0;
	}

        printd("MODE httpd access\n");

	ll->httpd_access_line = 1;
	if (reading_1st_loglines) 
	  mode_httpd_access = 1;
      }

    } else if (*s == '['			// [C] IS THIS AN APACHE ERROR LOG LINE ? ----------------------------------
          && *(s+4)  == ' '			// "[Wed May  9 22:09:18 2001] [notice] ..."
          && *(s+14) == ':'
          && *(s+25) == ']'
	  && *(s+26) == ' ') {

      printd1("MODE httpd apache errors0: %s\n", s);

      if (5 == sscanf(s + 5, "%3s %2d %d:%d:%d", month, &day, &hour, &minute, &second)) {
	ll->day		= (u_char) day;
	ll->hour	= (u_char) hour;
	ll->minute	= (u_char) minute;
	ll->second	= (u_char) second;
	strncpy(ll->month, month, 3);
	ll->month[3]	= '\0';

	skip_to_nth_c(s, 1, ']');		// skip the date
	s++;					// skip the blank
	ll->message = s;			// => this is the 'message'.

        printd("MODE httpd apache errors\n");

	ll->httpd_errors_line = 1;
	if (reading_1st_loglines) 
	  mode_httpd_errors = 1;
      }

    } else if (*s == '['			// [D] IS THIS A NETSCAPE SERVER ERROR LOG LINE ? --------------------------
          && *(s+3)  == '/'			// starts with "[31/Oct/2001:17:37:02] warning: ..."
          && *(s+12) == ':'
          && *(s+21) == ']'
	  && *(s+22) == ' ') {

      printd1("MODE httpd Netscape errors0: %s\n", s);

       if (6 == sscanf(s + 1, "%d/%3s/%d:%d:%d:%d",
                      &day, month, &year, &hour, &minute, &second)) {
	ll->day		= (u_char) day;
	ll->hour	= (u_char) hour;
	ll->minute	= (u_char) minute;
	ll->second	= (u_char) second;
	strncpy(ll->month, month, 4);

	skip_to_nth_c(s, 1, ']');		// skip the date
	s++;					// skip the blank
	ll->message = s;			// => this is the 'message'.

        printd("MODE httpd Netscape errors\n");

	ll->httpd_errors_line = 1;
	if (reading_1st_loglines) 
	  mode_httpd_errors = 1;
      }
    }
  }


  if (search_regexp)				// if a search was done, highlight matching
    search_logline(ll);				// substrings also in this new logline

  // fprintf(stderr, "LOGLINE 0x%x: %-100.100s\n", (int) ll, raw_line);

  return ll;
}


void initialize_line_numerotation()
{
  logline * ll;
  int       num = 0;
  int       nlines2print;

  ll = (new_tail_line ? new_tail_line : tail_line);
  if (ll && ll->linenum == MAXINT) {
    while (ll) {				// give decreasing linenumbers to loglines
      ll->linenum = num--;			// from bottom (line 0) to top

      if (! i_found_start_logline) {
        if ((option_tail_nnn_f > 0 || option_n_lines_only > 0)
         && ! option_start_month) {
          nlines2print = (option_tail_nnn_f > 0 ? option_tail_nnn_f : option_n_lines_only);
          if (ll->linenum == - nlines2print + 1) {
            cur_bottom_ll = ll;		// if user wants a given number of lines (options
            i_found_start_logline = 1;	// -<nn>f or -n <nn>), the start line is the one with
	    // fprintf(stderr, "START LINE: 0x%x %-90.90s\n", (int) ll, ll->line);
          }				// line number: - nlines2print + 1
        }
      }

      ll = ll->prev;
    }
  }
}


void check_for_start_line(
  logline * ll)
{
  int nlines2print;

  if (! i_found_start_logline) {	// if I haven't yet found the start line (if there's one)
    if (option_start_month) {		// if a start date was given on command line:
      if (ll->is_before_start_date)	// if this line is before start date, remember it
        i_saw_before_logline = 1;
      else if (i_saw_before_logline) {	// if this line is after start date, and if I saw
        cur_bottom_ll = ll;		// one before start date, tell main thread I saw
        i_found_start_logline = 1;	// the start date crossing.
      }

    } else if (option_tail_nnn_f > 0 || option_n_lines_only > 0) {
      nlines2print = (option_tail_nnn_f > 0 ? option_tail_nnn_f : option_n_lines_only);

      if (ll->linenum == - nlines2print + 1) {
        cur_bottom_ll = ll;		// if user wants a given number of lines (options
        i_found_start_logline = 1;	// -<nn>f or -n <nn>), the start line is the one with
      }					// line number -nlines2print + 1
    }
  }

  if (0 && cur_bottom_ll)
    fprintf(stderr, "LOGLINE 0x%x IS START LINE: %s\n", (int) ll, ll->line);
}

// -----------------------------------------------------------------------------------------------

/* append_buffer_to_loglines_queue()
 * 
 * append the lines contained in the passed buffer (null-terminated),
 * at the tail of the `struct logline' queue.
 * New pointer to end-of-queue is set in global variable new_tail_line.
 */

void append_buffer_to_loglines_queue(
  char * buffer)
{
  char *    eol;
  logline * ll = NULL;		// '= NULL' to avoid gcc warning
  char      my_oldest_date_seen[SMALLSTR];

//int       n = logspad_nlines / 2;

  *my_oldest_date_seen = '\0';
  i_saw_before_logline = 0;
  new_lines_arrived = 0;	// for main thread, in case previous new loglines are not yet processed
  new_tail_line = tail_line;	// to leave tail_line untouched, it is new_tail_line that
				// will walk to the new tail as I add loglines
  while (*buffer != '\0') {
    if ((eol = (char *) index(buffer, '\n')))
      *eol = '\0';

    if ((! *my_oldest_date_seen) && isdigit((int) buffer[5])	// keep date of first line in buffer
     && buffer[3] == ' ' && buffer[6] == ' ')			// (that has a valid date)
      sprintf(my_oldest_date_seen, "%6.6s", buffer);		// keep the date (for status displaying)

    if (((! option_i_ipmon_only) || strstr(buffer, "ipmon"))	// show only ipmon lines if option -i
     && (! option_g_grep					// if option -g <regexp>, check that
         || (grep_v_regexp					// the regexp matches (or doesn't match,
	     ^ (regexec(option_g_grep, buffer, 0, NULL, 0) == 0)))) {  // if option -v was given)

      // fprintf(stderr, "APPEND: %s.\n", buffer);

      ll = create_logline(buffer, NULL, new_tail_line);		// create a new `logline' to hold this line

      if (new_tail_line)		// if queue already exists,
        new_tail_line->next = ll;	//   then link this new line to the end of the queue
      else
        head_line = ll;			//   else initialize the head pointer
      new_tail_line = ll;		// this new logline is now the tail of the queue
      total_nbr_loglines_read++;

      check_for_start_line(ll);		// is this logline the start line I'm (maybe) looking for ?

/*
      if (! --n) {			// if we just added half a screenful of lines,
        new_lines_arrived = 1;		// pause till they're displayed
        while (new_lines_arrived)
          napms(200);
        n = logspad_nlines / 2;		// reinit countdown
      }
*/
    }

    if (eol)				// if eol is not null, there are other lines in buffer
      buffer = ++eol;			// skip the '\0' (that had replaced the original '\n')
    else				// and loop to add them
      break;
  }

  if (! tail_line)			// if I just created the queue (tail_line was NULL),
    initialize_line_numerotation();	// start line numbering

  if (*my_oldest_date_seen)				// if I saw a displayable logline with a valid date,
    strcpy(oldest_date_seen, my_oldest_date_seen);	// update the `oldest date seen'

  new_lines_arrived = 1;		// tell main thread that new lines arrived
}


/* prepend_buffer_to_loglines_queue()
 * 
 * prepend the lines contained in the passed buffer (null-terminated),
 * at the head of `struct logline' queue.
 * Update the pointer to head-of-queue (head_line).
 */

void prepend_buffer_to_loglines_queue(
  char * buffer)
{
  char *    eol;
  logline * sep_head = (logline *) NULL;
  logline * sep_tail = (logline *) NULL;
  logline * ll;
  char      my_oldest_date_seen[SMALLSTR];

  i_saw_before_logline = 0;
  new_backw_lines_read = 0;
  *my_oldest_date_seen = '\0';

  /* First create a new separate queue with the lines in buffer,
   * then link it at the head of the main loglines queue.
   */

  while (*buffer != '\0' && ! reverse_thread_must_stop) {
    if ((eol = (char *) index(buffer, '\n')))			// a '\n' in buffer ? replace it 
      *eol = '\0';						// with a '\0' to terminate one line.

    if ((! *my_oldest_date_seen) && isdigit((int) buffer[5])	// keep date of first line in buffer
     && buffer[3] == ' ' && buffer[6] == ' ')			// (that has a valid date)
      sprintf(my_oldest_date_seen, "%6.6s", buffer);		// keep the date (for status displaying)

    if (((! option_i_ipmon_only) || strstr(buffer, "ipmon"))	// show only ipmon lines if option -i
     && (! option_g_grep					// if option -g <regexp>, check that
         || (grep_v_regexp					// the regexp matches (or doesn't match,
	     ^ (regexec(option_g_grep, buffer, 0, NULL, 0) == 0)))) {  // if option -v was given)

      ll = create_logline(buffer, NULL, sep_tail);	// create a new `logline' to hold this line
      if (! sep_head)			// initialize sep_head if it is the 1st line
        sep_head = ll;			// of this separate queue
      if (sep_tail)			// link this new logline to the separate queue
        sep_tail->next = ll;		// unless this is the 1st line
      sep_tail = ll;			// anyway now the new tail is this very line
      total_nbr_loglines_read++;

      check_for_start_line(ll);		// is this logline the start line I'm (maybe) looking for ?
    }

    if (eol)				// if eol is not null, there are other lines
      buffer = ++eol;			// in buffer: skip the '\0' (that had replaced
    else				// the original '\n') and loop to add them
      break;
  }

  if (sep_head) {			// (if `ipmon lines only' or if option -g, we may have read no valid line)
    sep_tail->next = head_line;

    if (head_line) {			// if main queue exists (i.e. if /var/log/ipflog had valid lines in its tail),
      head_line->prev = sep_tail;	//   link the separate queue to main queue
      for (ll = sep_tail; ll; ll = ll->prev) {
        ll->linenum = ll->next->linenum - 1; // propagate line numbering
	check_for_start_line(ll);
      }

    } else {				// else
      tail_line = sep_tail;		//   initialize main queue tail pointer
      initialize_line_numerotation();	//   initialize line numbering
      screen_was_resized = 1;		//   artificial means to redraw screen, it was empty since
    }					//     loglines queue was empty
    head_line = sep_head;
  }

  if (*my_oldest_date_seen)				// if I saw a displayable logline with a valid date,
    strcpy(oldest_date_seen, my_oldest_date_seen);	// update the `oldest date seen'

  if (oldest_date_seen) {
    strcpy(reverse_thread_status, "Reading: ");		// set reverse-reading-thread status
    strcat(reverse_thread_status, oldest_date_seen);	// to last read line date: "Reading: Jun 23" 
  }
  new_backw_lines_read = 1;				// and tell main thread to update status line
}

