/*
 * Safer Memory Allocation
 * Copyright (C) 1997 Kaz Kylheku <kaz@cafe.net>
 *
 * Free Software License:
 *
 * All rights are reserved by the author, with the following exceptions:
 * Permission is granted to freely reproduce and distribute this software,
 * possibly in exchange for a fee, provided that this copyright notice appears
 * intact. Permission is also granted to adapt this software to produce
 * derivative works, as long as the modified versions carry this copyright
 * notice and additional notices stating that the work has been modified.
 * The copyright extends to translations of this work into other languages,
 * including machine languages. 
 *
 * $Id: smalloc.c,v 1.4 1997/08/23 14:42:45 kaz Exp $
 * $Name: release_1_4 $
 */

#undef NDEBUG
#include <assert.h>
#include <string.h>
#include <stdlib.h>
#include "smalloc.h"

/*
static char rcsid[] = "$Id: smalloc.c,v 1.4 1997/08/23 14:42:45 kaz Exp $";
static char right[] = "Copyright (C) 1997 Kaz Kylheku";
*/

#undef safe_malloc
#undef safe_calloc
#undef safe_free
#undef safe_realloc

static void setmagic(char *block, size_t size, const safe_magic_t *magic)
{
    memcpy(block + size, magic, sizeof *magic);
}

static int checkmagic(char *block, size_t size, const safe_magic_t *magic)
{
    return memcmp(block + size, magic, sizeof *magic) == 0;
}

static void markused(char *block, size_t size, int yes)
{
    block[size + sizeof(safe_magic_t)] = (yes != 0);
}

static int isused(char *block, size_t size)
{
    return block[size + sizeof(safe_magic_t)] == 1;
}

void *safe_malloc(size_t size, const safe_magic_t *magic)
{
    size_t size_adj = size + sizeof *magic + 1;
    void *ptr;

    assert (size != 0);
    assert (size_adj > size);

    ptr = malloc(size_adj);
    if (ptr) {
	setmagic(ptr, size, magic);
	markused(ptr, size, 1);
    }
    return ptr;
}

void *safe_calloc(size_t nmemb, size_t size, const safe_magic_t *magic)
{
    size_t total = nmemb * size;
    void *ptr;

    assert (total != 0);
    assert (total / size == nmemb);

    ptr = safe_malloc(total, magic);
    if (ptr)
	memset(ptr, 0, total);
    return ptr;
}

void safe_free(void *ptr, size_t size, const safe_magic_t *magic)
{
    assert (checkmagic(ptr, size, magic));
    assert (isused(ptr, size));
    markused(ptr, size, 0);
    free(ptr);
}

void *safe_realloc(void *oldptr, size_t oldsize, size_t newsize,
	const safe_magic_t *magic)
{
    size_t size_adj = newsize + sizeof *magic + 1;
    void *ptr;

    assert (size_adj > newsize);
    assert (oldptr == NULL || isused(oldptr, oldsize));
    assert (oldptr == NULL || checkmagic(oldptr, oldsize, magic));

    if (newsize == 0) {
	safe_free(oldptr, oldsize, magic);
	return NULL;
    }

    ptr = realloc(oldptr, size_adj);
    if (ptr) {
	setmagic(ptr, newsize, magic);
	markused(ptr, newsize, 1);
    } else if (newsize < oldsize) {
	setmagic(oldptr, newsize, magic);
	markused(oldptr, newsize, 1);
    }
    return ptr;
}
