/*
 * vlog.c
 *
 * vlog: curses based IP-Filter logs viewer
 *
 * Copyright (C) 2001  Pierre Berthomier
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2
 * as published by the Free Software Foundation, provided that this notice
 * is preserved in its entirety and due credit is given to the original
 * author and the contributors.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#define IPFLOG_MAIN
#include "vlog.h"


void usage()
{
  fprintf(stderr, "
%s v%s (%s)  %s
Curses based logfile viewer for IP-Filter

Options:
  -r                (rotated): read also old, rotated log files (logfile.0, ...)
  -i                show ipmon log lines only (default is to show all log lines)
  -g regexp         (grep) show only loglines that match 'regexp'
  -v regexp         (grep -v) show only loglines that do not match 'regexp'
  -l logfile        name of log file to be read
  -l{imsxladcq}     predefined logfile names: i = /var/log/ipflog (default),
                    m = /var/log/messages,    s = /var/log/syslog,
		    x = /var/log/xferlog,     l = /var/log/maillog,
		    a = /var/log/authlog,     d = /var/log/daemon,
		    c = /var/cron/log,        q = /var/log/squid.
  -<nn>f or -f      follow logfile: a simple `tail -<nn>f' (no curses)
  -<nn>  or -n <nn> print <nn> loglines on stdout and exit (no curses)
  -s [[Mmm] dd]     show lines from date Mmm dd hh:mm:ss. If Mmm and dd are
     [hh[:mm[:ss]]    not given, use current day. (ex: -s Jul 21) (ex: -s 12:30)
  -e ...            same thing as -s, specify end date (no curses)
  -o{dtuhfrilan}    choose which optional fields to display (date time usec
                      host intf rule isba-rule hdrlen, or all/none)
  -t                (text only) don't print ANSI codes (if no curses mode)
  -d                disguise IP addresses and hostnames for publishing purpose
  -w <nn> or -ww    force vlog output line width to nn chars max (if no curses)
  -h                this help                       (-ww = no width limitation)
", MYNAME, VERSION, DATE, CREDIT);

  exit(1);
}

/* Write the date in buf according to the format passed */

void localtime_to_buf(
  char * buf,
  char * format)
{
  time_t      curtime;
  struct tm * tm;

  time(&curtime);
  tm = localtime(&curtime);
  strftime(buf, 32, format, tm);
}

/* Returns the number of seconds since 00:00:00 of 1st day of current month */

long current_second()
{
  char buf[SMALLSTR];
  int  d, h, m, s;

  localtime_to_buf(buf, "%d %H %M %S");
  sscanf(buf, "%d %d %d %d", &d, &h, &m, &s);
  s += 60 * (m + 60 * (h + 24 * d));
  return s;
}


void parse_command_line(
  int argc,
  char **argv)
{
  char   c;
  char   buf[SMALLSTR];
  char * s;
  int    ret, figures, last_optind;

  mypathname = strdup(argv[0]);

  logfile = NULL;
  option_tail_nnn_f	= -1;
  option_n_lines_only	= -1;
  option_w_width	= 0;
  option_o_display	= 0;
  option_i_ipmon_only	= 0;
  option_r_rotated	= 0;
  option_g_grep		= NULL;
  option_t_text_only	= 0;
  grep_v_regexp		= 0;
  option_d_disguise_addr = 0;

  mode_ipmon		= 1;
  option_show_date      = 0;
  option_show_time      = 0;
  option_show_microsec  = 0;
  option_show_hostname  = 0;
  option_show_intf      = 0;
  option_show_rulenum   = 0;
  option_show_isbanum   = 0;
  option_show_headerlen = 0;
  option_show_proc_name = 0;
  option_show_pid	= 0;
  option_show_httpd_client_ip	= 0;
  option_show_httpd_client_user	= 0;
  option_show_httpd_code	= 0;
  option_show_httpd_length	= 0;

  option_start_month	= (char *) NULL;
  option_start_day	= -1;
  option_start_hour	= -1;
  option_start_minute	= -1;
  option_start_second	= -1;
  option_end_month	= (char *) NULL;
  option_end_day	= -1;
  option_end_hour	= -1;
  option_end_minute	= -1;
  option_end_second	= -1;
  mode_autoscroll	= 1;
  figures		= 0;
  last_optind		= 0;

  while ((c = getopt(argc, argv, "hf0123456789n:s:e:l:o:w:irdg:v:tT:")) != -1) {

    if ((c < '0' || c > '9') && c != 'f') {
      if (figures)
        option_n_lines_only = figures;
      figures = 0;
    }

    switch (c)
    {
      case 'h':
        usage();

      case 'f':							// -f
        option_tail_nnn_f = (figures ? figures : 10);
	figures = 0;
	break;

      case '0': case '1': case '2': case '3': case '4':
      case '5': case '6': case '7': case '8': case '9':		// -<nnn>f  or  -<nnn>
	/*
	fprintf(stderr, "arg = %s, next arg = %s, argc=%d, optind=%d\n",
	  argv[optind - 1],
	  (optind < argc ? argv[optind] : "NONE"),
	  argc, optind);
	*/

	figures = 10 * figures + (c - '0');
	break;

      case 'n':							// -n <nnn>
        option_n_lines_only = atoi(optarg);
        break;

      case 's':							// -s [[Mmm] dd] [hh[:mm[:ss]]]

	mode_autoscroll	= 0;			// don't auto scroll if a start date is given
	if (! isdigit((int) *optarg)) {		// -s Mmm ...
	  option_start_month = strdup(optarg);
	  if (optind >= argc || ! isdigit((int) *argv[optind])) // next arg begins with a digit ?
	    break;				// -s Mmm only.
	  s = argv[optind++];
	  option_start_day = atoi(s);		// -s Mmm dd.

	  if (optind >= argc || ! isdigit((int) *argv[optind])) // next arg begins with a digit ?
	    break;				// no, go on
	  s = argv[optind++];			// yes, put it in s and parse it below

	} else {				// -s dd ... or -s hh...
	  localtime_to_buf(buf, "%b");		// => start_month = current month (default)
	  option_start_month = strdup(buf);
	  s = optarg;
	  if (! strchr(s, ':')) {		// is there a ':' char ?
	    option_start_day = atoi(s);		// no. this a day number.
	    if (optind >= argc || ! strchr(argv[optind], ':')) // next arg contains a ':' ?
	      break;				// no, go on
	    s = argv[optind++];			// yes, it's a time spec.

	  } else {
	    localtime_to_buf(buf, "%d");	  // no day number specified
	    sscanf(buf, "%d", &option_start_day); // => start_day = current day
	  }
	}

	if (! isdigit((int) *s))		// Here s contains a time: hh[:mm[:ss]]
	  usage();
	sscanf(s, "%d", &option_start_hour);
	if ((s = strchr(s, ':'))) {
	  sscanf(++s, "%d", &option_start_minute);
	  if ((s = strchr(s, ':'))) {
	    sscanf(++s, "%d", &option_start_second);
	  }
	}
	if (option_start_hour   < -1 || option_start_hour   > 24
	 || option_start_minute < -1 || option_start_minute > 59
	 || option_start_second < -1 || option_start_second > 59)
	  usage();

        break;

      case 'e':							// -e [[Mmm] dd] [hh[:mm[:ss]]]

	if (! isdigit((int) *optarg)) {		// -s Mmm ...
	  option_end_month = strdup(optarg);
	  if (optind >= argc || ! isdigit((int) *argv[optind])) // next arg begins with a digit ?
	    break;				// -s Mmm only.
	  s = argv[optind++];
	  option_end_day = atoi(s);		// -s Mmm dd.

	  if (optind >= argc || ! isdigit((int) *argv[optind])) // next arg begins with a digit ?
	    break;				// no, go on
	  s = argv[optind++];			// yes, put it in s and parse it below

	} else {				// -s dd ... or -s hh...
	  localtime_to_buf(buf, "%b");		// => end_month = current month (default)
	  option_end_month = strdup(buf);
	  s = optarg;
	  if (! strchr(s, ':')) {		// is there a ':' char ?
	    option_end_day = atoi(s);		// no. this a day number.
	    if (optind >= argc || ! strchr(argv[optind], ':')) // next arg contains a ':' ?
	      break;				// no, go on
	    s = argv[optind++];			// yes, it's a time spec.

	  } else {
	    localtime_to_buf(buf, "%d");	  // no day number specified
	    sscanf(buf, "%d", &option_end_day); // => end_day = current day
	  }
	}

	if (! isdigit((int) *s))		// Here s contains a time: hh[:mm[:ss]]
	  usage();
	sscanf(s, "%d", &option_end_hour);
	if ((s = strchr(s, ':'))) {
	  sscanf(++s, "%d", &option_end_minute);
	  if ((s = strchr(s, ':'))) {
	    sscanf(++s, "%d", &option_end_second);
	  }
	}
	if (option_end_hour   < -1 || option_end_hour   > 24
	 || option_end_minute < -1 || option_end_minute > 59
	 || option_end_second < -1 || option_end_second > 59)
	  usage();

        break;

      case 'l':

	if (strlen(optarg) == 1) {
	  c = *optarg;
	  if (! (logfile = key_to_logfile[(int) c]))
	    usage();
	  if (c != 'i')
	    mode_ipmon = 0;

        } else {
	  logfile = (char *) strdup(optarg);
	  mode_ipmon = (strstr(logfile, "ipf") ? 1 : 0);	// go in ipmon mode if logfile
	}							// contains "ipf" string ...
	break;		// (readlog.c also goes to ipmon mode if first lines contains 'ipmon[' string)

      case 'o':
        for (s = optarg; *s; s++)
	  switch (*s) {
	    case 'd':	option_show_date = 1;		break;
	    case 't':	option_show_time = 1;		break;
	    case 'u':	option_show_microsec = 1; option_show_httpd_client_user = 1; break;
	    case 'h':	option_show_hostname = 1;	break;
	    case 'f':	option_show_intf = 1;		break;
	    case 'r':	option_show_rulenum = 1;	break;
	    case 'i':	option_show_isbanum = 1; option_show_pid = 1; break;
	    case 'l':	option_show_headerlen = 1;	break;
	    case 'p':	option_show_pid = 1;		break;
	    case 's':	option_show_httpd_client_ip = 1; break;
	    case 'c':	option_show_httpd_code = 1;	break;
	    case 'g':	option_show_httpd_length = 1;	break;
	    case 'a':	option_show_date        = option_show_time      = option_show_microsec =
	    		  option_show_hostname  = option_show_intf      = option_show_rulenum  =
	    		  option_show_isbanum   = option_show_headerlen =
			  option_show_proc_name = option_show_pid       =
			  option_show_httpd_code = option_show_httpd_length = 1;
	    		break;
	    case 'n':	option_show_date        = option_show_time      = option_show_microsec =
	    		  option_show_hostname  = option_show_intf      = option_show_rulenum  =
	    		  option_show_isbanum   = option_show_headerlen =
			  option_show_proc_name = option_show_pid       =
			  option_show_httpd_code = option_show_httpd_length = 0;
	    		break;
	    default:	usage();
	  }
	option_o_display = 1;
        break;

      case 'w':							// -w : force output to be w chars wide
        if (*optarg == 'w')					//      (-ww = wide width (show all line)
          option_w_width = 999;
	else
          option_w_width = atoi(optarg);
        break;

      case 'i':							// -i : show ipmon lines only
	option_i_ipmon_only = 1;
        break;

      case 'r':							// -r : read also rotated files (ipflog.0, ...)
        option_r_rotated = 1;
        break;

      case 'v':
        grep_v_regexp = 1;					// -v : "grep -v"

      case 'g':							// -g : "grep"
        option_g_grep = malloc(sizeof(regex_t));
	if ((ret = regcomp(option_g_grep, optarg, REG_EXTENDED)))
	  err("regexp is malformed (code=%d)", ret);
        break;

      case 't':							// -t : text only (no bold chars)
        option_t_text_only = 1;
        break;
/*
      case 'T':							// -T <term-name>
        sprintf(buf, "TERM=%s", optarg);
	putenv(buf);
        break;
*/
      case 'd':							// -H: hide IP addresses and
        option_d_disguise_addr = 				//     first letters of hostnames
	  (int) ((long) current_second() & 0xffffffff);		//     (current_second is used as a random number)
        break;

      default:
        usage();
    }
  }

  if (figures)
    option_n_lines_only = figures;

  if (! logfile)
    logfile = (char *) strdup("/var/log/ipflog");

  if (option_start_month) {
    option_start_month[0] = ucase(option_start_month[0]);	// enforce first letter is upper case,
    option_start_month[1] = lcase(option_start_month[1]);	// other are lower
    option_start_month[2] = lcase(option_start_month[2]);

    if (month_to_num(option_start_month) < 0)			// is that a month name I know ?
      exit(1);							// if not, a message was printed. exit.
  }

  if (option_end_month) {
    option_end_month[0] = ucase(option_end_month[0]);		// enforce first letter is upper case,
    option_end_month[1] = lcase(option_end_month[1]);		// other are lower
    option_end_month[2] = lcase(option_end_month[2]);

    if (month_to_num(option_end_month) < 0)			// is that a month name I know ?
      exit(1);							// if not, a message was printed. exit.
    if (! option_start_month)
      err("when an end date is specified, a start date should be too");
  }

}


void msg(char * string, ...)
{
  va_list arglist;
  va_start(arglist, string);
  fprintf(stderr, "%s: ", MYNAME);
  vfprintf(stderr, string, arglist);
  fprintf(stderr, "\n");
  va_end(arglist);
}

void err(char * string, ...)
{
  va_list arglist;

  flushinp();
  if (mode_curses) endwin();
  va_start(arglist, string);
  fprintf(stderr, "%s: ", MYNAME);
  vfprintf(stderr, string, arglist);
  fprintf(stderr, "\n");
  va_end(arglist);
  exit(1);
}

void err_perr(char * string, ...)
{
  va_list arglist;

  flushinp();
  if (mode_curses) endwin();
  va_start(arglist, string);
  fprintf(stderr, "%s: ", MYNAME);
  vfprintf(stderr, string, arglist);
  va_end(arglist);
  fprintf(stderr, ": ");
  perror("");
  exit(1);
}

void * tmalloc(int nbytes)		// "threaded" malloc
{					// called by concurrent threads => mutex required
  void * p;
  
  pthread_mutex_lock(&malloc_mutex);	// get the lock
  p = (void *) malloc(nbytes);		// allocate memory
  if (!p)
    err_perr("malloc failed");
  pthread_mutex_unlock(&malloc_mutex);	// release the lock

  return p;
}

void tfree(void * p)			// "threaded" free
{
  if (p) {
    pthread_mutex_lock(&malloc_mutex);
    free(p);
    pthread_mutex_unlock(&malloc_mutex);
  } else
    err("tfree(): trying to free a null pointer ?!?");
}

char * tstrdup(char * s)		// "threaded" strdup
{
  char * t = NULL;

  if (s) {
    pthread_mutex_lock(&malloc_mutex);
    t = strdup(s);
    pthread_mutex_unlock(&malloc_mutex);
  }

  return t;
}

void sleep_us(int microsec)
{
  struct timeval tv; 
  fd_set         fs;

  tv.tv_sec = 0;
  tv.tv_usec = microsec;
  FD_ZERO(&fs);
  FD_SET(0, &fs);
  select(1, &fs, NULL, NULL, &tv);
}


int month_to_num(
  char * month)
{
  if (! strncmp(month, "Jan", 3)) return 0;
  if (! strncmp(month, "Feb", 3)) return 1;
  if (! strncmp(month, "Mar", 3)) return 2;
  if (! strncmp(month, "Apr", 3)) return 3;
  if (! strncmp(month, "May", 3)) return 4;
  if (! strncmp(month, "Jun", 3)) return 5;
  if (! strncmp(month, "Jul", 3)) return 6;
  if (! strncmp(month, "Aug", 3)) return 7;
  if (! strncmp(month, "Sep", 3)) return 8;
  if (! strncmp(month, "Oct", 3)) return 9;
  if (! strncmp(month, "Nov", 3)) return 10;
  if (! strncmp(month, "Dec", 3)) return 11;

  transient_message("?! invalid month specification (%s)", month);
  return -1;
}

// month1 and month2 are tested on the first three letters (Jan, Feb, Mar, ...)
// Return -1 if month1 is one of the six months BEFORE month2
// return +1 if month1 is one of the six months AFTER month2
// return  0 if month1 == month2.

int compare_months(
  char * month1,
  char * month2)
{
  int i1, i2, diff;

  i1 = month_to_num(month1);
  i2 = month_to_num(month2);

  if (i1 >= 0 && i2 >= 0) {
    diff = i2 - i1;
    if (diff > 5)
      diff = diff - 12;
    else if (diff < -5)
      diff = diff + 12;

    return (diff > 0 ? 1 : (diff < 0 ? -1 : 0));
  }

  return 0;	// error in month name ... return "equal" ...
}


// Return -1 if logline ll is older than the date/time passed,
// return +1 if it is more recent, or 0 if it is equal.
// The date/time passed may be incomplete: month [day [hour [minute [second]]]]
// (day, hour, minute, second may be equal to -1 == unspecified)

int compare_ll_date(
  logline * ll,
  char *    month,
  int       day,
  int       hour,
  int       minute,
  int       second)
{
  int ret;

  if ((ret = compare_months(ll->month, month)))
    return ret;

  if (day >= 0) {
    if (ll->day != day)
      return (ll->day < day ? -1 : +1);

    if (hour >= 0) {
      if (ll->hour != hour)
        return (ll->hour < hour ? -1 : +1);

      if (minute >= 0) {
	if (ll->minute != minute)
          return (ll->minute < minute ? -1 : +1);

	if (second >= 0) {
	  if (ll->second != second)
            return (ll->second < second ? -1 : +1);
        }
      }
    }
  }

  return 0;
}


void fill_my_attr_table()
{
  int i;

  for (i = 0; i < 256; i++) {
    my_attr_table[i] = (chtype)
     (((i & MYA_NORMAL)		? A_NORMAL	: 0)
    | ((i & MYA_BOLD)		? A_BOLD	: 0)
    | ((i & MYA_REVERSE)	? A_REVERSE	: 0)
    | ((i & MYA_UNDERLINE)	? A_UNDERLINE	: 0)
    | ((i & MYA_DIM)		? A_DIM		: 0))
    ^ ((i & MY_SEARCH_MATCH)	? A_REVERSE	: 0);	// chars matching a search are reversed (xor)
  }
}

// return the number of "ESC [" sequences in buf till 'maxwidth' characters.

int count_nbr_of_ansi_codes_till_maxwidth(
  char * buf,
  int    maxwidth)
{
  int i = 0;

  while (*buf && maxwidth > 0) {
    if (*buf == '' && *(buf + 1) == '[') {
      i++;
      maxwidth += 4;
    }
    maxwidth--;
    buf++;
  }

  return i;
}


/* -------------------------------------------------------------------------- */

void init_data()
{
  int i, ret;
  char home_vlogrc[BUFSIZ];

  head_line		= (logline *) NULL;
  tail_line		= (logline *) NULL;
  new_tail_line		= (logline *) NULL;
  headpad		= (WINDOW *)  NULL;
  logspad		= (WINDOW *)  NULL;
  statswin		= (WINDOW *)  NULL;
  horiz_scroll		= 0;
  screen_maxx		= 0;

  screen_was_resized	= 0;
  new_lines_arrived	= 0;
  new_backw_lines_read	= 0;
  old_logfiles_fd	= NULL;

  nchars_host   	= 11;
  cur_bottom_ll		= NULL;
  last_ll		= NULL;
  i_am_at_top		= 0;
  total_nbr_loglines_read = 0;
  i_found_start_logline = 0;
  reverse_thread_must_stop = 0;
  proc_name_width	= 10;
  reading_1st_loglines = 1;

  mode_raw_cooked	= 1;
  mode_curses		= 0;
  mode_httpd_access	= 0;
  mode_httpd_errors	= 0;

  option_sparse_ip_addr = 0;
  search_regexp		= (regex_t *) NULL;
  cur_fmt_serial	= 1;
  cur_srch_serial	= 1;
  search_status		= SRCH_NOP;
  search_thread_stop	= 0;
  isba_rulenums_available = 0;
  debug_int		= 0;
  last_typed_key_second	= current_second();

  str_to_chstr(blank_line, A_NORMAL, "%*s", PADWIDTH, "");
  gethostname(local_host_name, SMALLSTR);

  ret = pthread_mutex_init(&malloc_mutex, 0);
  if (ret)
    err("pthread_mutex_init() failed (error %d)", ret);

  if (! getenv("TERM"))
    putenv("TERM=vt100");

  fill_my_attr_table();

  for (i = 0; i < 256; i++)
    key_to_logfile[i] = (char *) NULL;

  sprintf(home_vlogrc, "%s/.vlogrc", getenv("HOME"));
  set_predefined_key_to_logfile_map();
  read_vlog_conf_file("/etc/vlog.conf");
  read_vlog_conf_file(home_vlogrc);
}

void set_predefined_key_to_logfile_map()
{
  struct stat statbuf;

  key_to_logfile['i'] = (char *) strdup("/var/log/ipflog");
  key_to_logfile['s'] = (char *) strdup("/var/log/syslog");
  key_to_logfile['d'] = (char *) strdup("/var/log/daemon");
  key_to_logfile['l'] = (char *) strdup("/var/log/maillog");
  key_to_logfile['x'] = (char *) strdup("/var/log/xferlog");
  key_to_logfile['a'] = (char *) strdup("/var/log/authlog");
  key_to_logfile['q'] = (char *) strdup("/var/log/squid");
  key_to_logfile['n'] = (char *) strdup("/var/log/snort/alert");
  key_to_logfile['p'] = (char *) strdup("/var/log/snort/portscan.log");

  if (! stat("/var/cron/olog" , &statbuf))			// 'c' = /var/cron/olog if it exists,
    key_to_logfile['c'] = (char *) strdup("/var/cron/olog");	//       /var/cron/log otherwise
  else
    key_to_logfile['c'] = (char *) strdup("/var/cron/log");

  if (! stat("/var/adm/messages" , &statbuf))			// 'm' = /var/adm/messages if it exists,
    key_to_logfile['m'] = (char *) strdup("/var/adm/messages");	//       /var/log/messages otherwise
  else
    key_to_logfile['m'] = (char *) strdup("/var/log/messages");
}

void read_vlog_conf_file(
  char *  vlog_conf_file)
{
  FILE * fd;
  char   buf[BUFSIZ];
  char   logfile[BUFSIZ];
  char * s;
  char   key;

  if ((fd = fopen(vlog_conf_file, "r"))) {
    while (fgets(buf, BUFSIZ, fd)) {
      if ((s = index(buf, '#')))
        *s = '\0';
      s = buf;
      if (*s) {
        if (2 == sscanf(s, "key2logfile %c %s", &key, logfile)) {
	  key_to_logfile[(int) key] = strdup(logfile);
	  // fprintf(stderr, "%s: %c -> %s\n", vlog_conf_file, key, logfile);
        }
      }
    }
  }
}


void no_curses_get_screen_maxx()
{
  struct winsize wsize;
  int tty_fildes;

  if (option_w_width)			// if user gave "-w <nn>", use that value
    screen_maxx = option_w_width;

  else if (((tty_fildes = open("/dev/tty", O_RDWR, 0777)) >= 0)
        && (ioctl(tty_fildes, TIOCGWINSZ, &wsize) == 0)) {
    close(tty_fildes);
    screen_maxx = (int) wsize.ws_col;

  } else
    screen_maxx = 80;
}

void init_what_to_display()
{
  if (! mode_curses)			// if we're not in curses mode,
    no_curses_get_screen_maxx();	// I must seek screen_maxx myself

  if (!option_o_display) {		// if first call, and if display options not
    if (mode_ipmon) {			// chosen on command line, preset what to display
      option_show_time      = 1;	// according to window width

      if (screen_maxx > 100)
        option_show_intf      = 1;

      if (screen_maxx > 120) {
        option_show_date      = 1;
        option_show_rulenum   = 1 - isba_rulenums_available;
        option_show_isbanum   = isba_rulenums_available;
      }

      if (screen_maxx > 140) {
        option_show_microsec  = 0;
        option_show_hostname  = 1;
        option_show_rulenum   = 1;
        option_show_isbanum   = isba_rulenums_available;
        option_show_headerlen = 1;
      }

    } else {				// if (! mode_ipmon)
      option_show_date      = (screen_maxx <  90 ? 0 : 1);
      option_show_time      = 1;
      option_show_hostname  = (screen_maxx < 130 ? 0 : 1);
      option_show_proc_name = 1;
      option_show_pid       = (screen_maxx < 110 ? 0 : 1);
      option_show_httpd_client_ip   = 1;
      option_show_httpd_code        = (screen_maxx <  90 ? 0 : 1);
      option_show_httpd_length      = (screen_maxx < 110 ? 0 : 1);
      option_show_httpd_client_user = (screen_maxx < 130 ? 0 : 1);
    }
  }
}

/* -------------------------------------------------------------------------- */

int main(int argc, char ** argv)
{
  int  draw_1st_backlines_flag = 1;
  int  nbr_lines_printed;
  int  nbr_ansi_codes;
  char buffer[65536];

  // tests();

  open_debugfile();
  init_data();
  parse_command_line(argc, argv);
  if (mode_ipmon)
    init_ipf_module();
  start_logfiles_reading_threads(logfile);

  if (option_tail_nnn_f < 0     	// ===== CURSES MODE ==================================================
   && option_n_lines_only < 0
   && ! option_end_month) {

    init_curses();
    init_signal_handlers();
    init_what_to_display();
    draw_head_pad();
    draw_status_window();
    refresh_screen();

    while (1) {
      move(0, screen_maxx - 1);
      napms(MAIN_LOOP_MS);
      react_to_typed_keys();

      if (cur_bottom_ll) {			// if cur_bottom_ll is null, maybe we have not yet seen
						// the start line (because of option -s)
        if (i_found_start_logline) {		// if I just found the start line,
          option_start_month = NULL;		// reset flags, I don't need them anymore
	  i_found_start_logline = 0;
	  cur_bottom_ll->is_marked = 1;		// mark start logline
	  napms(250);				// wait for some preceeding loglines to be read
	  ll_walk_back(cur_bottom_ll, logspad_nlines / 2);	// (for if I'm near the top)
	  ll_walk_forward(cur_bottom_ll, logspad_nlines);	// display start line in middle of screen
        }							// go on. below, we'll draw screen.

        if (! option_start_month) {		// if a start date was given, don't display anything
      						// until I've found the start logline (then option_start_month = 0)
          if (new_lines_arrived) {		// some new log lines were read and added at tail
                                 		// by the "tail -f" thread
            if (mode_autoscroll) {		// if 'auto-scrolldown' mode is enabled:
              if (cur_bottom_ll != tail_line)	// 1. if we're not at bottom, remember where we were
                last_ll = cur_bottom_ll;
              cur_bottom_ll = new_tail_line;	// 2. go to bottom and scroll automatically (tail -f)
              redraw_screen();
            }
            tail_line = new_tail_line;
	    new_tail_line = NULL;
            new_lines_arrived = 0;
          }

          if (screen_was_resized)		// screen_was_resized is set by sigwinch_handler()
            hard_redraw_screen();

          if (new_backw_lines_read) {		// new_backw_lines_read is set by reverse_reading_thread()
            if (draw_1st_backlines_flag) {
              redraw_screen();
	      draw_1st_backlines_flag = 0;
            }
            new_backw_lines_read = 0;
          }

          if (search_status & SRCH_ENDED) {
            redraw_screen();
            search_status &= ~SRCH_ENDED;
          }
        }
      }

      draw_status_window();			// refresh the "Reading: XXX" and time info
      wnoutrefresh(statswin);
      doupdate();
    }

  } else {	// ===== NO CURSES MODE ===================================================================

    init_what_to_display();
    nbr_lines_printed = 0;

    while (1) {
      if (cur_bottom_ll) {			// if cur_bottom_ll is null, maybe we have not yet seen
						// the start line (because of option -s)
	while (1) {
	  if (! cur_bottom_ll->was_printed) {
	    log_line_to_fmt_and_attr_lines(cur_bottom_ll);

	    if (option_end_month		// if an end date was given: have I crossed it ?
	     && compare_ll_date(cur_bottom_ll, option_end_month, option_end_day,
		  option_end_hour, option_end_minute, option_end_second) == +1) {
	      exit(0);				// yes, exit.
	    }

	    if (screen_maxx != 999) {				// print log line.
	      if (option_t_text_only)
	        printf("%-*.*s\n", screen_maxx, screen_maxx, cur_bottom_ll->fmtline);

	      else {
	        fmt_and_attr_lines_to_ANSI_string(cur_bottom_ll, buffer);
	        nbr_ansi_codes = count_nbr_of_ansi_codes_till_maxwidth(buffer, screen_maxx);
		//fprintf(stderr, "nbr_ansi_codes = %d\n", nbr_ansi_codes);

	        printf("%-*.*s%s\n",
	          screen_maxx + 4 * nbr_ansi_codes,
		  screen_maxx + 4 * nbr_ansi_codes,
		  buffer,
		  ANSI_NORMAL);
	      }

	    } else						// if option "-ww" (print full logline)
	      printf("%s\n",
	        (option_t_text_only
	         ? cur_bottom_ll->fmtline
	         : fmt_and_attr_lines_to_ANSI_string(cur_bottom_ll, buffer)));

	    cur_bottom_ll->was_printed = 1;

	    if (option_n_lines_only > 0) {			// if option -n <nn>: have I displayed the
	      if (++nbr_lines_printed >= option_n_lines_only	// specified nbr of lines ?
	       || cur_bottom_ll == tail_line) {			// or have I reached the end of logfile ?
	        while (old_logfiles_fd)				// wait for the reverse thread pipe to be closed
	          napms(50);					// to avoid the 'Broken pipe' message
	        exit(0);
	      }
	    }
	  }
	  if (cur_bottom_ll->next)
	    cur_bottom_ll = cur_bottom_ll->next;
	  else
	    break;
	}
      }

      if (new_lines_arrived) {		// some new log lines were read and added at tail
        tail_line = new_tail_line;	// => i must update tail_line.
        new_tail_line = NULL;
        new_lines_arrived = 0;
      }

      napms(500);
    }
  }

  exit(0);
}


