/*
 * vlog.h
 *
 * vlog: curses based IP-Filter logs viewer
 *
 * Copyright (C) 2001  Pierre Berthomier
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2
 * as published by the Free Software Foundation, provided that this notice
 * is preserved in its entirety and due credit is given to the original
 * author and the contributors.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

// TODO:
//
// o append_buffer_to_loglines_queue():
//			test: flow it (very fast loglines arrival)
// o ipfmodule.c	- print my own IP addresses in bold chars
//			- kill -HUP -> reread ifconfig -a, ipfstat -ion, ipf.conf
//			- parse 'frag', NAT and STATE lines
//
// BUGS:
//   searching on big logfiles is not good
//   ./vlog -e Jul 30 -s Jun 29   fails
//

// XFree86:
//   PgUp   = 0x153
//   PgDown = 0x152
//   Home   = 0x106
//   End    = 0x168
//   BS     = 0x107
//   Ins    = 0x14b (Solaris: 7e)
//   Suppr  = 0x07f (Solaris: 7f)

#ifndef __Solaris__
#define __Solaris__ (defined(sun) && (defined(__svr4__) || defined(__SVR4)))
#endif

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <unistd.h>
#include <string.h>
#include <strings.h>
#include <errno.h>
#include <signal.h>
#include <pthread.h>
#include <stdarg.h>		// for msg(varargs ...)
#include <curses.h>
#include <regex.h>
#include <time.h>
#include <fcntl.h>
#include <libgen.h>		// for basename()
#include <sys/types.h>
#include <sys/stat.h>
#include <netinet/in.h>		// for protocol number definitions (IPPROTO_TCP ...)

#ifdef __FreeBSD__
#  include <sys/ttycom.h>
#endif
#ifdef __OpenBSD__
#  include <sys/ioctl.h>
#  include <sys/ttycom.h>
#endif
#ifdef __Solaris__
#  include <sys/termios.h>
#endif

#include "hash.h"


#define MYNAME		"vlog"
#define VERSION		"1.1f"
#define DATE		"08 Fev 02"
#define	CREDIT		"Pierre Berthomier (c) 2001-2002"
#define MAXINT		((u_int) (1 << 31) - 1)
#define LASTNBYTES	32768		// #bytes to read 1st from end of logfile at start
#define PADWIDTH	512		// should be set to max width of a raw log line
#define SMALLSTR	256
#define IPF_PASS	0
#define IPF_BLOCK	1
#define PROTO_NAME_MAX	256		// size of protocols names array
#define SERV_NAME_MAX	16384		// size of tcp and udp services names arrays

#define LLFLAG_NONE		 0
#define LLFLAG_MARKED		 1
#define LLFLAG_REGEXP_MATCH	 2
#define LLFLAG_LOGFILE_SEPARATOR 4
#define LLFLAG_NOT_AN_IPMON_LINE 8

#define	SRCH_NOP		0
#define SRCH_FOUND_ONE		1
#define SRCH_IN_PROGRESS	2
#define SRCH_BACKWARDS		4
#define SRCH_ENDED		8

#define AUTO_REDRAW_DELAY		60	// redraw screen every 60 sec when no keys are hit
#define MAIN_LOOP_MS			100	// delay in wait-for-keys loop (millisec)
#define TAIL_F_THREAD_SLEEP_MS		500
#define BACKWARDS_THREAD_SLEEP_MS	100
#define TAIL_F_READ_BLOCK		65536	// tail -f reading thread: read block size
#define BACKWARDS_READ_BLOCK		1048576	// backwards reading thread: read block size
						// block sizes must be > ipfline maxwidth (256)
#define CHK_ROTATED_LOGFILE_COUNT	20	// loop count to check if logfile rotated

#define SCROLL_UP	0
#define SCROLL_RIGHT	1
#define SCROLL_DOWN	2
#define SCROLL_LEFT	3

#define MYA_NORMAL	0		// compact the char attributes into 1 byte (instead of 3 bytes)
#define MYA_BOLD	1
#define MYA_REVERSE	2
#define MYA_UNDERLINE	4
#define MYA_DIM		8
#define MY_SEARCH_MATCH	64		// chars matching last regexp search have this bit equal to 1
#define MY_BIT7		128		// all attr chars have this set => attrs strings can be manipulated
					// as a null-terminated string
#define ANSI_NORMAL	"[0m"		// ANSI control chars for no-curses modes
#define ANSI_BOLD	"[1m"
#define ANSI_UNDERLINE	"[4m"
#define ANSI_REVERSE	"[7m"

#define LINTF			14
#define HTTP_CLIENT_IP_NCOLS	15
#define HTTP_CLIENT_USER_NCOLS	10

#define max(a, b)	((a) > (b) ? (a) : (b))
#define min(a, b)	((a) < (b) ? (a) : (b))

/* ---------------------------------------------------------------------------------------- */


typedef struct logline_fields_s {		// ipf fields (for loglines generated by ipmon)
  int     microsec;
  u_short mult;		/* logline multiplier ("12x"), or 1 if none	*/
  char    intf[LINTF];	/* intf name OR STATE:* OR NAT:*	 	*/

  u_char  sa1;		/* sa1-4 = src address dec numbers */
  u_char  sa2;
  u_char  sa3;
  u_char  sa4;
  u_char  da1;		/* da1-4 = dst address dec numbers */
  u_char  da2;
  u_char  da3;
  u_char  da4;
  u_short sp;		/* sp = source port if applicable */
  u_short dp;		/* dp = destination port if applicable */

  short   isbarule;	/* rule number in isba source ruleset, if possible 	*/
  short   groupnum;	/* '10' extracted from @10:25				*/
  short   rulenum;	/* '25' extracted from @10:25 (or from '@25 NAT:*')	*/
  u_char  protocol; 	/* protocol number */
  u_char  headerlen;	/* IP header length */
  int     packetlen;	/* packet length */

  u_short
	  unknown_format:1,	/* all structure is meaningless is this bit is 1 */
	  passblock:1,		/* passed/blocked (0/1) */
	  in_out:1,		/* IN/OUT (0/1) */
	  keep_state:1,		/* unused if not TCP/UDP/ICMP */
	  keep_frags:1,		/* unused if not TCP */
	  tcp_S:1,
	  tcp_A:1,
	  tcp_P:1,
	  tcp_U:1,
	  tcp_F:1,
	  tcp_R:1;

  u_char  icmp_type;	/* unused if not ICMP */
  u_char  icmp_code;

//int      expire_npkts;  /* for NAT:EXPIRE, STATE:EXPIRE, STATE:CLOSE */
//uint64_t expire_nbytes;

  char * icmp_for;	/* if ICMP packet "for", descr. of orig pkt */

} logline_fields;

/* ---------------------------------------------------------------------------------------- */

typedef struct logline_s {
  char *	     line;		// log line raw string
  short					// log line flags:
		     is_marked:1,		// true if this logline was marked by user
		     matched:1,			// true if it matches current regexp (see also cur_srch_serial)
		     is_a_separator:1,		// true if this 'logline' actually contains a log file name
		     not_an_ipmon_line:1,	// true if this logline is not an ipmon logline
		     httpd_access_line:1,	// true if this logline is an HTTPD access file logline
		     httpd_errors_line:1,	// true if this logline is an HTTPD errors file logline
		     emitted_by_local_host:1,	// true if this logline was emitted by local host
		     is_before_start_date:1,	// true if this logline's date is older that start date
		     was_printed:1;		// no curses mode: true if this logline was printed on screen

  u_char	     serial;		// incremented each time user changes display format
  u_char	     srch_serial;	// incremented each time user gives a new regexp to search
  struct logline_s * next;		// next     logline in queue
  struct logline_s * prev;		// previous logline in queue
  int		     linenum;		// line 0 = last logline at the time when vlog is launched

  char *	     host;		// remember here some logline properties common to all logfiles
  char		     month[4];
  u_char	     day;
  u_char	     hour;
  u_char	     minute;
  u_char	     second;
  int		     pid;		// pid of process that generated message (for non-ipmon syslog lines)
  char *             message;		// message body (in case of plain, non-ipmon syslog line)
					// (points somewhere in self->line)
  char *             client_addr;	// for Apache log lines: IP addr or FQDN of client
  char *             client_user;	// for Apache log lines: user name in case there was an authentication
  int                http_code;		// for Apache log lines: HTTP return code
  int                http_size;		// for Apache log lines: HTTP transfert size
  logline_fields *   fields;		// struct of fields extracted from logline (intf, src addr, ...)
  char *	     fmtline;		// formatted log line string
  char *	     attrline;		// and its attributes, compacted in a byte (MYA_*)
} logline;

/* ---------------------------------------------------------------------------------------- */

#ifdef  IPFLOG_MAIN
#  define EXTERN
#else
#  define EXTERN extern
#endif

EXTERN char    * logfile;
EXTERN logline * head_line;
EXTERN logline * tail_line;
EXTERN logline * new_tail_line;	// new tail when the "tail -f" thread read new log lines
EXTERN logline * fhead_line;
EXTERN logline * ftail_line;
EXTERN logline * cur_bottom_ll;	// points to the (logline *) currently displayed at bottom of win
EXTERN logline * last_ll;	// remember the line we were on before typing 'G'
EXTERN pthread_t tail_f_thread_id;
EXTERN pthread_t reverse_reading_thread_id;
EXTERN pthread_t search_thread_id;
EXTERN pthread_mutex_t malloc_mutex;
EXTERN long	 youngest_logfile_start_offset;
EXTERN FILE *    old_logfiles_fd;
EXTERN int	 number_of_oldest_logfile;
EXTERN int	 i_am_at_top;
EXTERN int	 total_nbr_loglines_read;

EXTERN int	 screen_maxx;
EXTERN int	 screen_maxy;
EXTERN int	 logspad_nlines;
EXTERN int	 screen_was_resized;
EXTERN int	 new_lines_arrived;
EXTERN int	 new_backw_lines_read;
EXTERN WINDOW *  headpad;	// Ipf fields header pad (1 line)    | use pads to be able
EXTERN WINDOW *  logspad;	// middle pad for log lines          | to scroll to the right
EXTERN WINDOW *  statswin;	// bottom status window (1 line)
EXTERN int	 horiz_scroll;	// nb of chars scrolled to left (for headpad and logspad)
EXTERN u_char	 cur_fmt_serial;// incremented each time user changes log lines display format
				//   (used to flag fmtlines with a != serial, as "dirty")

EXTERN regex_t * search_regexp;
EXTERN int	 grep_v_regexp;		// is TRUE if option '-v' was given (grep -v)
EXTERN u_char	 cur_srch_serial; 	// incremented each time user gives a new regexp to search
EXTERN int	 search_status;
EXTERN logline * search_result_ll;
EXTERN int	 search_nbr_matches;
EXTERN int	 search_thread_stop;	// set to 1 by main thread for search thread to stop
EXTERN int	 nchars_host;		// hostname field width in logs window

EXTERN int	 mode_raw_cooked;	// 1 = loglines are displayed cooked, 0 = raw
EXTERN int	 mode_autoscroll;	// 1 = go to log bottom when a new log line arrives
EXTERN int	 mode_ipmon;		// 1 = we parse ipmon log lines, 0 = any other format (syslogd)
EXTERN int	 mode_httpd_access;	// 1 = we parse an HTTPD access or errors logfile
EXTERN int	 mode_httpd_errors;	// 1 = we parse an HTTPD access or errors logfile
EXTERN int       mode_curses;		// 1 = curses are used (initscr() was called);
EXTERN int	 option_show_date;
EXTERN int	 option_show_time;
EXTERN int	 option_show_microsec;
EXTERN int	 option_show_hostname;
EXTERN int	 option_show_intf;
EXTERN int	 option_show_rulenum;
EXTERN int	 option_show_isbanum;
EXTERN int	 option_show_headerlen;
EXTERN int	 option_show_proc_name;
EXTERN int	 option_show_pid;
EXTERN int	 option_show_httpd_client_ip;
EXTERN int	 option_show_httpd_client_user;
EXTERN int	 option_show_httpd_code;
EXTERN int	 option_show_httpd_length;

EXTERN chtype	 my_attr_table[256];
EXTERN char      local_host_name[SMALLSTR];

EXTERN int	 isba_rulenums_available;
EXTERN int	 option_sparse_ip_addr;
EXTERN chtype    blank_line[PADWIDTH + 1];
EXTERN char	 reverse_thread_status[SMALLSTR];
EXTERN int	 reverse_thread_must_stop;

EXTERN int	option_tail_nnn_f;
EXTERN int	option_n_lines_only;
EXTERN int	option_w_width;
EXTERN int	option_o_display;
EXTERN int	option_i_ipmon_only;
EXTERN int	option_r_rotated;
EXTERN regex_t * option_g_grep;
EXTERN int	option_t_text_only;
EXTERN int	option_d_disguise_addr;

EXTERN char *	option_start_month;
EXTERN int	option_start_day;
EXTERN int	option_start_hour;
EXTERN int	option_start_minute;
EXTERN int	option_start_second;
EXTERN int	i_found_start_logline;
EXTERN int	i_saw_before_logline;

EXTERN char *	option_end_month;
EXTERN int	option_end_day;
EXTERN int	option_end_hour;
EXTERN int	option_end_minute;
EXTERN int	option_end_second;

EXTERN int	last_typed_key_second;
EXTERN int	debug_int;
EXTERN char *	mypathname;
EXTERN char *	key_to_logfile[256];
EXTERN int	proc_name_width;	// in non-ipmon mode, nbr of chars for process name column
EXTERN int	reading_1st_loglines;

/* ---------------------------------------------------------------------------------------- */

/* vlog.c */

void		localtime_to_buf(char *, char *);
long		current_second();
void		msg(char *, ...);
void		err(char *, ...);
void		err_perr(char *, ...);
void		sleep_us(int);
void *		tmalloc(int);
void		tfree(void *);
char *		tstrdup(char *);
void		usage();
int		month_to_num(char *);
int		compare_months(char *, char *);
int		compare_ll_date(logline *, char *, int, int, int, int);
void		set_predefined_key_to_logfile_map();
void		read_vlog_conf_file(char *);

/* readlog.c */

void		start_logfiles_reading_threads(char *);
void		tail_f_thread(FILE *);
void		reverse_reading_thread(long *);
logline *	create_logline(char *, logline *, logline *);
void		initialize_line_numerotation();
void		append_buffer_to_loglines_queue(char *);
void		prepend_buffer_to_loglines_queue(char *);

/* curses.c */

void		init_curses();
void		init_signal_handlers();
void		sigwinch_handler(int);
void		create_windows();
int		chstrlen(chtype *);
chtype *	str_to_chstr(chtype *, int, char *, ...);
chtype *	strcat_to_chstr(chtype *, int, char *, ...);
char *		create_const_attrline(int, char);
void		transient_message(char *, ...);
char *		fmt_and_attr_lines_to_ANSI_string(logline *, char *);

void		help_screen();
void		draw_head_pad();
chtype *	headers_append_host_logfilename(chtype *);
void		log_line_to_fmt_and_attr_lines(logline *);
void		draw_loglines_pad();
void		draw_status_window();
void		clear_status_window();
void		refresh_screen();
void		redraw_screen();
void		hard_redraw_screen();
void		clean_curses_and_exit(int);

int		search_logline(logline *);
void		search_thread(int);
void		get_search_regexp(int);
void		get_what_to_toggle();
void		scroll_screen(int, int);
void		go_to_top_of_logs();
void		react_to_typed_keys();


/* ipfmodule.c */

void		init_ipf_module();
void		cut_logline_to_ipfilter_fields(logline *);
void		ipmon_fields_to_fmt_and_attr_lines(logline * ll);
chtype *	ipf_headers_line(chtype *);


/* syslogmodule.c */

void		syslog_fields_to_fmt_and_attr_lines(logline * ll);
chtype *	syslog_headers_line(chtype *);


/* httpdmodule.c */

void		httpd_fields_to_fmt_and_attr_lines(logline * ll);
chtype *	httpd_headers_line(chtype *);


// -------------------------------------------------------------------------------------
// MACROS


/* Copy chars from src to dst until the nth occurence of char c (excluded)
 * or until a '\0' char in src.
 * Return a pointer to *two* chars after the last one copied from src (skip the char c)
 * dst is terminated by a '\0'.
 * dst is supposed to have enough free memory ahead.
 *
 * src must be a (char *), dst may be a (char[]).
 */

#define copy_to_nth_c(src, dst, n, c) {	\
  char * _dst = dst;			\
  int    _n = n;			\
  while (*(src)) {			\
    if (*(src) == c)			\
      if (--_n <= 0)			\
        break;				\
    *_dst++ = *(src)++;			\
  }					\
  *_dst = '\0';				\
  (src)++;				\
}

/* Same thing, with a limitation on length of string copied (ncmax - 1 chars copied max) */

#define ncopy_to_nth_c(src, dst, n, c, ncmax) {	\
  char * _dst = dst;			\
  int    _n = n;			\
  int    _ncmax = ncmax;		\
  while (*(src) && --_ncmax) {	\
    if (*(src) == c)			\
      if (--_n <= 0)			\
        break;				\
    *_dst++ = *(src)++;			\
  }					\
  *_dst = '\0';				\
  (src)++;				\
}

/* Skip to the char after the nth char c
 */

#define skip_to_nth_c(src, n, c) {	\
  int _n = n;				\
  while (*(src)) {			\
    if (*(src)++ == c)			\
      if (--_n <= 0)			\
        break;				\
  }					\
}

#define nappend_to_fmt_and_attr_strings(fmt_str, attr_str, str, myattr, n) {	\
  int _n = n;									\
  char * s = str;								\
  while (*s && _n--) {								\
    *fmt_str++ = *s++;								\
    *attr_str++ = myattr;							\
  }										\
}

#define append_to_fmt_and_attr_strings(fmt_str, attr_str, str, myattr) {	\
  char * s = str;								\
  while (*s) {									\
    *fmt_str++ = *s++;								\
    *attr_str++ = myattr;							\
  }										\
}

/* Convert a chstr (chtype *) to a standard str (char *)
 */

#define chstr_to_str(chstr, str) {	\
  chtype * cs = chstr;			\
  char   * s  = str;			\
  while (*cs)				\
    *s++ = (char) (*cs++ & 0xff);	\
}

/* logical XOR
 */

#define XOR(a, b)			\
  (((a) && ! (b)) || ((b) && ! (a)))


/* Make logline pointer ll go back 'nlines' lines if possible
 * (if there are enough lines before ll)
 */

#define ll_walk_back(ll, nlines) {	\
  int i = nlines;			\
  while (ll->prev && i-- > 0)		\
    ll = ll->prev;			\
}

/* Make logline pointer ll go forward 'nlines' lines if possible
 * (if there are enough lines after ll)
 */

#define ll_walk_forward(ll, nlines) {	\
  int i = nlines;			\
  while (ll->next && i-- > 0)		\
    ll = ll->next;			\
}

#define strcat_and_advance_pointer(str, str2) {	\
  char * _str2 = str2;				\
  while (*_str2)				\
    *str++ = *_str2++;				\
}

/* Transform lower case letter to upper case. Other chars are unchanged */

#define ucase(c)			\
  ((c) >= 'a' && (c) <= 'z' ? (char) ((c) - 'a' + 'A') : (c))

/* Transform upper case letter to lower case. Other chars are unchanged */

#define lcase(c)			\
  ((c) >= 'A' && (c) <= 'Z' ? (char) ((c) - 'A' + 'a') : (c))

/* debugging */

#ifdef DEBUG

EXTERN FILE * fd_debug;
#  define open_debugfile()			fd_debug = fopen("/tmp/vd", "w")
#  define close_debugfile()			fclose(fd_debug)
#  define debug_timestamp() {			\
     char buf[64];				\
     localtime_to_buf(buf, "%H:%M:%S ");	\
     fprintf(fd_debug, buf);			\
   }
#  define printd(format)			fprintf(fd_debug, format);		     fflush(fd_debug);
#  define printd1(format, arg1)			fprintf(fd_debug, format, arg1);	     fflush(fd_debug);
#  define printd2(format, arg1, arg2)		fprintf(fd_debug, format, arg1, arg2);	     fflush(fd_debug);
#  define printd3(format, arg1, arg2, arg3)	fprintf(fd_debug, format, arg1, arg2, arg3); fflush(fd_debug);

#else
#  define open_debugfile()
#  define close_debugfile()
#  define debug_timestamp()
#  define printd(format)
#  define printd1(format, arg1)
#  define printd2(format, arg1, arg2)
#  define printd3(format, arg1, arg2, arg3)
#endif

